# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.io.pdbx"
__author__ = "Patrick Kunzmann"
__all__ = ["PDBxFile"]

import copy
import shlex
from collections.abc import MutableMapping
import numpy as np
from ....file import TextFile


class PDBxFile(TextFile, MutableMapping):
    """
    This class represents a PDBx/mmCIF file.

    The categories of the file can be accessed using the
    :meth:`get_category()`/:meth:`set_category()` methods.
    The content of each category is represented by a dictionary.
    The dictionary contains the entry
    (e.g. *label_entity_id* in *atom_site*) as key.
    The corresponding values are either strings in *non-looped*
    categories, or 1-D numpy arrays of string objects in case of
    *looped* categories.

    A category can be changed or added using :meth:`set_category()`:
    If a string-valued dictionary is provided, a *non-looped* category
    will be created; if an array-valued dictionary is given, a
    *looped* category will be created. In case of arrays, it is
    important that all arrays have the same size.

    Alternatively, The content of this file can also be read/write
    accessed using dictionary-like indexing:
    You can either provide a data block and a category or only a
    category, in which case the first data block is taken.

    Notes
    -----
    This class is also able to detect and parse multiline entries in the
    file. However, when writing a category no multiline values are used.
    This could lead to long lines.

    This class uses a lazy category dictionary creation: When reading
    the file only the line positions of all categories are checked. The
    time consuming task of dictionary creation is done when
    :meth:`get_category()` is called.

    Examples
    --------
    Read the file and get author names:

    >>> import os.path
    >>> file = PDBxFile.read(os.path.join(path_to_structures, "1l2y.cif"))
    >>> author_dict = file.get_category("citation_author", block="1L2Y")
    >>> print(author_dict["name"])
    ['Neidigh, J.W.' 'Fesinmeyer, R.M.' 'Andersen, N.H.']

    Dictionary style indexing, no specification of data block:

    >>> print(file["citation_author"]["name"])
    ['Neidigh, J.W.' 'Fesinmeyer, R.M.' 'Andersen, N.H.']

    Get the structure from the file:

    >>> arr = get_structure(file)
    >>> print(type(arr).__name__)
    AtomArrayStack
    >>> arr = get_structure(file, model=1)
    >>> print(type(arr).__name__)
    AtomArray

    Modify atom array and write it back into the file:

    >>> arr_mod = rotate(arr, [1,2,3])
    >>> set_structure(file, arr_mod)
    >>> file.write(os.path.join(path_to_directory, "1l2y_mod.cif"))
    """

    def __init__(self):
        super().__init__()
        # This dictionary saves the PDBx category names,
        # together with its line position in the file
        # and the data_block it is in
        self._categories = {}

    @classmethod
    def read(cls, file):
        """
        Read a PDBx/mmCIF file.

        Parameters
        ----------
        file : file-like object or str
            The file to be read.
            Alternatively a file path can be supplied.

        Returns
        -------
        file_object : PDBxFile
            The parsed file.
        """
        file = super().read(file)
        # Remove emptyline at then end of file, if present
        if file.lines[-1] == "":
            del file.lines[-1]

        current_category = None
        start = -1
        stop = -1
        is_loop = False
        has_multiline_values = False
        for i, line in enumerate(file.lines):
            # Ignore empty and comment lines
            if not _is_empty(line):
                data_block_name = _data_block_name(line)
                if data_block_name is not None:
                    data_block = data_block_name
                    # If new data block begins, reset category data
                    current_category = None
                    start = -1
                    stop = -1
                    is_loop = False
                    has_multiline_values = False

                is_loop_in_line = _is_loop_start(line)
                category_in_line = _get_category_name(line)
                if is_loop_in_line or (
                    category_in_line != current_category
                    and category_in_line is not None
                ):
                    # Start of a new category
                    # Add an entry into the dictionary with the old category
                    stop = i
                    file._add_category(
                        data_block,
                        current_category,
                        start,
                        stop,
                        is_loop,
                        has_multiline_values,
                    )
                    # Track the new category
                    if is_loop_in_line:
                        # In case of lines with "loop_" the category is in the
                        # next line
                        category_in_line = _get_category_name(
                            file.lines[i + 1]
                        )
                    is_loop = is_loop_in_line
                    current_category = category_in_line
                    start = i
                    has_multiline_values = False

                multiline = _is_multi(line, is_loop)
                if multiline:
                    has_multiline_values = True
        # Add the entry for the final category
        # Since at the end of the file the end of the category
        # is not determined by the start of a new one,
        # this needs to be handled separately
        stop = len(file.lines)
        file._add_category(
            data_block,
            current_category,
            start,
            stop,
            is_loop,
            has_multiline_values,
        )
        return file

    def get_block_names(self):
        """
        Get the names of all data blocks in the file.

        Returns
        -------
        blocks : list
            List of data block names.
        """
        blocks = set()
        for category_tuple in self._categories.keys():
            block, _ = category_tuple
            blocks.add(block)
        return sorted(blocks)

    def get_category(self, category, block=None, expect_looped=False):
        """
        Get the dictionary for a given category.

        Parameters
        ----------
        category : string
            The name of the category. The leading underscore is omitted.
        block : string, optional
            The name of the data block. Default is the first
            (and most times only) data block of the file.
        expect_looped : bool, optional
            If set to true, the returned dictionary will always contain
            arrays (only if the category exists):
            If the category is *non-looped*, each array will contain
            only one element.

        Returns
        -------
        category_dict : dict of (str or ndarray, dtype=str) or None
            A entry keyed dictionary. The corresponding values are
            strings or array of strings for *non-looped* and
            *looped* categories, respectively.
            Returns None, if the data block does not contain the given
            category.
        """
        if block is None:
            block = self.get_block_names()[0]
        category_info = self._categories.get((block, category))
        if category_info is None:
            return None
        start = category_info["start"]
        stop = category_info["stop"]
        is_loop = category_info["loop"]
        is_multilined = category_info["multiline"]

        if is_multilined:
            # Convert multiline values into singleline values
            prelines = [
                line.strip()
                for line in self.lines[start:stop]
                if not _is_empty(line) and not _is_loop_start(line)
            ]
            lines = (len(prelines)) * [None]
            # lines index
            k = 0
            # prelines index
            i = 0
            while i < len(prelines):
                if prelines[i][0] == ";":
                    # multiline values
                    multi_line_str = prelines[i][1:]
                    j = i + 1
                    while prelines[j] != ";":
                        multi_line_str += prelines[j]
                        j += 1
                    lines[k - 1] += " " + shlex.quote(multi_line_str)
                    i = j + 1
                elif not is_loop and prelines[i][0] in ["'", '"']:
                    # Singleline values where value is in the line
                    # after the corresponding key
                    lines[k - 1] += " " + prelines[i]
                    i += 1
                else:
                    # Normal singleline value in the same row as the key
                    lines[k] = prelines[i]
                    i += 1
                    k += 1
            lines = [line for line in lines if line is not None]

        else:
            lines = [
                line.strip()
                for line in self.lines[start:stop]
                if not _is_empty(line) and not _is_loop_start(line)
            ]

        if is_loop:
            # Special optimization for "atom_site":
            # Even if the values are quote protected,
            # no whitespace is expected in escaped values
            # Therefore slow shlex.split() call is not necessary
            if category == "atom_site":
                whitespace_values = False
            else:
                whitespace_values = True
            category_dict = _process_looped(lines, whitespace_values)
        else:
            category_dict = _process_singlevalued(lines)

        if expect_looped and not is_loop:
            category_dict = {
                key: np.array([val], dtype=object)
                for key, val in category_dict.items()
            }

        return category_dict

    def set_category(self, category, category_dict, block=None):
        """
        Set the content of a category.

        If the category is already exisiting, all lines corresponding
        to the category are replaced. Otherwise a new category is
        created and the lines are appended at the end of the data block.

        Parameters
        ----------
        category : string
            The name of the category. The leading underscore is omitted.
        category_dict : dict
            The category content. The dictionary must have strings
            (subcategories) as keys and strings or :class:`ndarray`
            objects as values.
        block : string, optional
            The name of the data block. Default is the first
            (and most times only) data block of the file. If the
            block is not contained in the file yet, a new block is
            appended at the end of the file.
        """
        if block is None:
            block = self.get_block_names()[0]

        # Determine whether the category is a looped category
        sample_category_value = list(category_dict.values())[0]
        if isinstance(sample_category_value, (np.ndarray, list)):
            is_looped = True
            # Check whether all arrays have the same length
            arr_len = len(list(category_dict.values())[0])
            for subcat, array in category_dict.items():
                if len(array) != arr_len:
                    raise ValueError(
                        f"Length of Subcategory '{subcat}' is {len(array)}, "
                        f" but {arr_len} was expected"
                    )
        else:
            is_looped = False

        # Sanitize dictionary
        # -> convert to string
        # -> replace empty values with '.'
        category_dict = copy.deepcopy(category_dict)
        if is_looped:
            for subcat, value in category_dict.items():
                array = np.asarray(value)
                # Cast array if its data type is not a Unicode string
                if array.dtype.kind != "U":
                    array = array.astype(str)
                array = np.char.strip(array)
                array[array == ""] = "."
                category_dict[subcat] = array
        else:
            for subcat, value in category_dict.items():
                value = str(value)
                value = value if value != "" else "."
                category_dict[subcat] = str(value)

        # Value arrays (looped categories) can be modified (e.g. quoted)
        # Hence make a copy to avoid unwanted side effects
        # due to modification of input values
        if is_looped:
            category_dict = {
                key: val.copy() for key, val in category_dict.items()
            }

        # Enclose values with quotes if required
        for key, value in category_dict.items():
            if is_looped:
                # Since value is a numpy string array with fixed size,
                # we need to convert it as a list before using _quote
                category_dict[key] = np.asarray(
                    [_quote(item) for item in value.tolist()]
                )
            else:
                category_dict[key] = _quote(value)

        if is_looped:
            keylines = [
                "_" + category + "." + key + " "
                for key in category_dict.keys()
            ]
            value_arr = list(category_dict.values())
            # Array containing the number of characters + whitespace
            # of each column
            col_lens = np.zeros(len(value_arr), dtype=int)
            for i, column in enumerate(value_arr):
                col_len = 0
                for value in column:
                    if len(value) > col_len:
                        col_len = len(value)
                # Length of column is max value length
                # +1 whitespace character as separator
                col_lens[i] = col_len + 1
            valuelines = [""] * arr_len
            for i in range(arr_len):
                for j, arr in enumerate(value_arr):
                    valuelines[i] += arr[i] + " " * (col_lens[j] - len(arr[i]))
            newlines = ["loop_"] + keylines + valuelines

        else:
            # For better readability, not only one space is inserted
            # after each key, but as much spaces that every value starts
            # at the same position in the line
            max_len = 0
            for key in category_dict.keys():
                if len(key) > max_len:
                    max_len = len(key)
            # "+3" Because of three whitespace chars after longest key
            req_len = max_len + 3
            newlines = [
                "_" + category + "." + key + " " * (req_len - len(key)) + value
                for key, value in category_dict.items()
            ]

        # A comment line is set after every category
        newlines += ["#"]

        if (block, category) in self._categories:
            # Category already exists in data block
            category_info = self._categories[(block, category)]
            # Insertion point of new lines
            old_category_start = category_info["start"]
            old_category_stop = category_info["stop"]
            category_start = old_category_start
            # Difference between number of lines of the old and new category
            len_diff = len(newlines) - (old_category_stop - old_category_start)
            # Remove old category content
            del self.lines[old_category_start:old_category_stop]
            # Insert new lines at category start
            self.lines[category_start:category_start] = newlines
            # Update category info
            category_info["start"] = category_start
            category_info["stop"] = category_start + len(newlines)
            # When writing a category no multiline values are used
            category_info["multiline"] = False
            category_info["loop"] = is_looped
        elif block in self.get_block_names():
            # Data block exists but not the category
            # Find last category in the block
            # and set start of new category to stop of last category
            last_stop = 0
            for category_tuple, category_info in self._categories.items():
                if block == category_tuple[0]:
                    if last_stop < category_info["stop"]:
                        last_stop = category_info["stop"]
            category_start = last_stop
            category_stop = category_start + len(newlines)
            len_diff = len(newlines)
            self.lines[category_start:category_start] = newlines
            self._add_category(
                block,
                category,
                category_start,
                category_stop,
                is_looped,
                is_multilined=False,
            )
        else:
            # The data block does not exist
            # Put the begin of data block in front of newlines
            newlines = ["data_" + block, "#"] + newlines
            # Find last category in the file
            # and set start of new data_block with new category
            # to stop of last category
            last_stop = 0
            for category_info in self._categories.values():
                if last_stop < category_info["stop"]:
                    last_stop = category_info["stop"]
            category_start = last_stop + 2
            category_stop = last_stop + len(newlines)
            len_diff = len(newlines) - 2
            self.lines[last_stop:last_stop] = newlines
            self._add_category(
                block,
                category,
                category_start,
                category_stop,
                is_looped,
                is_multilined=False,
            )
        # Update start and stop of all categories appearing after the
        # changed/added category
        for category_info in self._categories.values():
            if category_info["start"] > category_start:
                category_info["start"] += len_diff
                category_info["stop"] += len_diff

    def __copy_fill__(self, clone):
        super().__copy_fill__(clone)
        clone._categories = copy.deepcopy(self._categories)

    def __setitem__(self, index, item):
        block, category_name = self._full_index(index)
        self.set_category(category_name, item, block=block)

    def __getitem__(self, index):
        block, category_name = self._full_index(index)
        return self.get_category(category_name, block=block)

    def __delitem__(self, index):
        block, category_name = self._full_index(index)
        category_info = self._categories[(block, category_name)]
        # Insertion point of new lines
        category_start = category_info["start"]
        category_stop = category_info["stop"]
        del self.lines[category_start:category_stop]
        # Update start and stop of all categories appearing after the
        # deleted category
        len_diff = category_stop - category_start
        for category_info in self._categories.values():
            if category_info["start"] > category_start:
                category_info["start"] -= len_diff
                category_info["stop"] -= len_diff

    def __contains__(self, index):
        block, category_name = self._full_index(index)
        return (block, category_name) in self._categories

    def __iter__(self):
        return self._categories.__iter__()

    def __len__(self):
        return len(self._categories)

    def _full_index(self, index):
        """
        Converts a an integer or tuple index into a block and a category
        name.
        """
        if isinstance(index, tuple):
            return index[0], index[1]
        elif isinstance(index, str):
            return self.get_block_names()[0], index
        else:
            raise TypeError(
                f"'{type(index).__name__}' is an invalid index type"
            )

    def _add_category(
        self, block, category_name, start, stop, is_loop, is_multilined
    ):
        # Before the first category starts,
        # the current_category is None
        # This is checked before adding an entry
        if category_name is not None:
            self._categories[(block, category_name)] = {
                "start": start,
                "stop": stop,
                "loop": is_loop,
                "multiline": is_multilined,
            }


def _process_singlevalued(lines):
    category_dict = {}
    i = 0
    while i < len(lines):
        parts = shlex.split(lines[i])
        key = parts[0].split(".")[1]
        if len(parts) > 1:
            value = parts[1]
        else:
            # The value is not in the same line,
            # but in the following one
            i += 1
            value = shlex.split(lines[i])[0]
        category_dict[key] = value
        i += 1
    return category_dict


def _process_looped(lines, whitepace_values):
    category_dict = {}
    keys = []
    # Array index
    i = 0
    # Dictionary key index
    j = 0
    for line in lines:
        if line[0] == "_":
            # Key line
            key = line.split(".")[1]
            keys.append(key)
            # Pessimistic array allocation
            # numpy array filled with strings
            category_dict[key] = np.zeros(len(lines), dtype=object)
            keys_length = len(keys)
        else:
            # If whitespace is expected in quote protected values,
            # use standard shlex split
            # Otherwise use much more faster whitespace split
            # and quote removal if applicable,
            # bypassing the slow shlex module
            if whitepace_values:
                values = shlex.split(line)
            else:
                values = line.split()
                for k in range(len(values)):
                    # Remove quotes
                    if (values[k][0] == '"' and values[k][-1] == '"') or (
                        values[k][0] == "'" and values[k][-1] == "'"
                    ):
                        values[k] = values[k][1:-1]
            for value in values:
                category_dict[keys[j]][i] = value
                j += 1
                if j == keys_length:
                    # If all keys have been filled with a value,
                    # restart with first key with incremented index
                    j = 0
                    i += 1
    for key in category_dict.keys():
        # Trim to correct size
        category_dict[key] = category_dict[key][:i]
    return category_dict


def _is_empty(line):
    return len(line) == 0 or line[0] == "#"


def _data_block_name(line):
    if line.startswith("data_"):
        return line[5:]
    else:
        return None


def _is_loop_start(line):
    return line.startswith("loop_")


def _is_multi(line, is_loop):
    if is_loop:
        return line[0] == ";"
    else:
        return line[0] in [";", "'", '"']


def _get_category_name(line):
    if line[0] != "_":
        return None
    else:
        return line[1 : line.find(".")]


def _quote(value):
    if "'" in value:
        return '"' + value + '"'
    elif '"' in value:
        return "'" + value + "'"
    elif " " in value:
        return "'" + value + "'"
    else:
        return value



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This subpackage provides functions and data structures for creating
(phylogenetic) trees.

The :class:`Tree` is the central class in this subpackage.
It wraps a *root* :class:`TreeNode` object.
A :class:`TreeNode` is either an intermediate node, if it has child
:class:`TreeNode` objects, or otherwise a leaf node.

A :class:`Tree` is not a container itself:
Objects, e.g species names or sequences, that are represented by the
nodes, cannot be stored directly in a :class:`Tree` or
:class:`TreeNode`.
Instead, each leaf node has a reference index:
These indices refer to a separate list or array, containing the actual
reference objects.

A :class:`Tree` can be created from or exported to a *Newick* notation,
usingthe :func:`Tree.from_newick()` or :func:`Tree.to_newick()` method,
respectively.

A :class:`Tree` can be build from a pairwise distance matrix using the
popular *UPGMA* (:func:`upgma()`) and *Neighbor-Joining*
(:func:`neighbor_joining()`) algorithms.
"""

__name__ = "biotite.sequence.phylo"
__author__ = "Patrick Kunzmann"

from .tree import *
from .upgma import *
from .nj import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.io.gff"
__author__ = "Patrick Kunzmann"
__all__ = ["get_annotation", "set_annotation"]

from ...annotation import Location, Feature, Annotation


def get_annotation(gff_file):
    """
    Parse a GFF3 file into an :class:`Annotation`.

    The *type* column is used as the :attr:`Feature.key` attribute,
    the locations (``loc``) are taken from the *start*, *end* and
    *strand* columns and  the *attributes* column is parsed into the
    :attr:`Feature.qual` attribute.
    Multiple entries with the same ``ID`` attribute are interpreted
    as the same feature.
    Thus, for entries with the same ``ID``, the *type* and *attributes*
    are only parsed once and the locations are aggregated from each
    entry.
    
    Parameters
    ----------
    gff_file : GFFFile
        The file tro extract the :class:`Annotation` object from.
    
    Returns
    -------
    annotation : Annotation
        The extracted annotation.
    """
    annot = Annotation()
    current_key = None
    current_locs = None
    current_qual = None
    current_id = None
    for _, _, type, start, end, _, strand, _, attrib in gff_file:
        id = attrib.get("ID")
        if id != current_id or id is None:
            # current_key is None, when there is no previous feature
            # (beginning of the file)
            if current_key is not None:
                # Beginning of new feature -> Save previous feature
                annot.add_feature(
                    Feature(current_key, current_locs, current_qual)
                )
            # Track new feature
            current_key = type
            current_locs = [Location(start, end, strand)]
            current_qual = attrib
        else:
            current_locs.append(Location(start, end, strand))
        current_id = id
    # Save last feature
    if current_key is not None:
        annot.add_feature(Feature(current_key, current_locs, current_qual))
    return annot


def set_annotation(gff_file, annotation,
                   seqid=None, source=None, is_stranded=True):
    """
    Write an :class:`Annotation` object into a GFF3 file.

    Each feature will get one entry for each location it has.
    :class:`Feature` objects with multiple locations require the ``ID``
    qualifier in its :attr:`Feature.qual` attribute.
    
    Parameters
    ----------
    gff_file : GFFFile
        The GFF3 file to write into.
    annotation : Annotation
        The annoation which is written to the GFF3 file.
    seqid : str, optional
        The content for the *seqid* column.
    source : str, optional
        The content for the *source* column.
    is_stranded : bool, optional
        If true, the strand of each feature is taken into account.
        Otherwise the *strand* column is filled with '``.``'.
    """
    for feature in sorted(annotation):
        if len(feature.locs) > 1 and "ID" not in feature.qual:
            raise ValueError(
                "The 'Id' qualifier is required "
                "for features with multiple locations"
            )
        ## seqid ##
        if seqid is not None and " " in seqid:
            raise ValueError("The 'seqid' must not contain whitespaces")
        ## source ##
        #Nothing to be done
        ## type ##
        type = feature.key
        ## strand ##
        # Expect same strandedness for all locations
        strand = list(feature.locs)[0].strand if is_stranded else None
        ## score ##
        score = None
        ## attributes ##
        attributes = feature.qual
        # The previous properties are shared by all entries
        # for this feature
        # The following loop handles properties that change with each
        # location
        reverse_order = True if strand == Location.Strand.REVERSE else False
        next_phase = 0
        for loc in sorted(
            feature.locs, key=lambda loc: loc.first, reverse=reverse_order
        ):
            ## start ##
            start = loc.first
            ## end ##
            end = loc.last
            ## strand ##
            strand = loc.strand if is_stranded else None
            ## phase ##
            if type == "CDS":
                phase = next_phase
                # Subtract the length of the location
                next_phase -= loc.last - loc.first + 1
                next_phase %= 3
            else:
                phase = None
            gff_file.append(
                seqid, source, type, start, end,
                score, strand, phase, attributes
            )


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import itertools
import numpy as np
import pytest
import biotite.sequence as seq
import biotite.sequence.align as align
import biotite.application.muscle as muscle
from biotite.application import VersionError
from ...util import is_not_installed
from .util import sequences


def test_align_ungapped():
    """
    Test `align_ungapped()` function via a known test case.
    """
    seq1 = seq.NucleotideSequence("ACCTGA")
    seq2 = seq.NucleotideSequence("ACTGGT")
    matrix = align.SubstitutionMatrix.std_nucleotide_matrix()
    ali = align.align_ungapped(seq1, seq2, matrix)
    assert ali.score == 3
    assert str(ali) == "ACCTGA\nACTGGT"


# [local, gap_penalty, input1, input2, expect]
align_cases = [(False,True, -7,      "TATGGGTATCC","TATGTATAA",
                    ("TATGGGTATCC\nTATG--TATAA",
                     "TATGGGTATCC\nTAT-G-TATAA",
                     "TATGGGTATCC\nTAT--GTATAA",)),
               (True, True, -6,      "TATGGGTATCC","TATGTATAA",
                    ("TATGGGTAT\nTATG--TAT",
                     "TATGGGTAT\nTAT-G-TAT",
                     "TATGGGTAT\nTAT--GTAT",)),
               (False,True, (-7,-1), "TACTATGGGTATCC","TCATATGTATAA",
                    ("TACTATGGGTATCC\nTCATATG--TATAA",
                     "TACTATGGGTATCC\nTCATAT--GTATAA",)),
               (True, True, (-7,-1), "TACTATGGGTATCC","TCATATGTATAA",
                    ("TATGGGTAT\nTATG--TAT",
                     "TATGGGTAT\nTAT--GTAT",)),
               (False,True, (-7,-1), "T","TTT",
                    ("T--\nTTT",
                     "--T\nTTT",)),
               (False,True, -7, "TAAAGCGAAAT","TGCGT",
                    ("TAAAGCGAAAT\nT---GCG---T")),
               (False,False,-7, "TAAAGCGAAAT","TGCGT",
                    ("TAAAGCGAAAT\n---TGCGT---"))
              ]
@pytest.mark.parametrize("local, term, gap_penalty, input1, input2, expect",
                         align_cases)
def test_align_optimal_simple(local, term, gap_penalty,
                              input1, input2, expect):
    """
    Test `align_optimal()` function using constructed test cases.
    """
    seq1 = seq.NucleotideSequence(input1)
    seq2 = seq.NucleotideSequence(input2)
    matrix = align.SubstitutionMatrix.std_nucleotide_matrix()
    # Test alignment function
    alignments = align.align_optimal(seq1, seq2,
                       matrix,
                       gap_penalty=gap_penalty, terminal_penalty=term,
                       local=local)
    
    for ali in alignments:
        assert str(ali) in expect
    # Test if separate score function calculates the same score
    for ali in alignments:
        score = align.score(ali, matrix,
                            gap_penalty=gap_penalty, terminal_penalty=term)
        assert score == ali.score


@pytest.mark.skipif(
    is_not_installed("muscle"),
    reason="MUSCLE is not installed"
)
# Ignore warning about MUSCLE writing no second guide tree
@pytest.mark.filterwarnings("ignore")
@pytest.mark.parametrize("gap_penalty, seq_indices", itertools.product(
    [-10, (-10,-1)], [(i,j) for i in range(10) for j in range(i+1)]
))
def test_align_optimal_complex(sequences, gap_penalty, seq_indices):
    """
    Test `align_optimal()` function using real world sequences,
    compared to the output of MUSCLE.
    """
    matrix = align.SubstitutionMatrix.std_protein_matrix()
    index1, index2 = seq_indices
    seq1 = sequences[index1]
    seq2 = sequences[index2]
    test_alignment = align.align_optimal(
        seq1, seq2, matrix,
        gap_penalty=gap_penalty, terminal_penalty=True, max_number=1
    )[0]

    try:
        ref_alignment = muscle.MuscleApp.align(
            [seq1, seq2], matrix=matrix, gap_penalty=gap_penalty
        )
    except VersionError:
        pytest.skip(f"Invalid Muscle software version")

    # Check whether the score of the optimal alignments is the same
    # or higher as the MUSCLE alignment
    # Direct alignment comparison is not feasible,
    # since the treatment of terminal gaps is different in MUSCLE
    test_score = align.score(
        test_alignment, matrix, gap_penalty, terminal_penalty=True
    )
    ref_score = align.score(
        ref_alignment, matrix, gap_penalty, terminal_penalty=True
    )
    try:
        assert test_score >= ref_score
    except AssertionError:
        print("Alignment:")
        print()
        print(test_alignment)
        print("\n")
        print("Reference alignment:")
        print()
        print(ref_alignment)
        raise


@pytest.mark.parametrize(
    "local, term, gap_penalty, seed", itertools.product(
        [True, False], [True, False], [-5, -8, -10, -15], range(10)
    )
)
def test_affine_gap_penalty(local, term, gap_penalty, seed):
    """
    Expect the same alignment results for a linear gap penalty and an
    affine gap penalty with the same gap open and extension penalty.
    """
    LENGTH_RANGE = (10, 100)
    MAX_NUMBER = 1000

    np.random.seed(seed)
    sequences = []
    for _ in range(2):
        sequence = seq.NucleotideSequence()
        length = np.random.randint(*LENGTH_RANGE)
        sequence.code = np.random.randint(
            len(sequence.alphabet), size=length
        )
        sequences.append(sequence)
    
    matrix = align.SubstitutionMatrix.std_nucleotide_matrix()

    ref_alignments = align.align_optimal(
        *sequences, matrix, gap_penalty, term, local, MAX_NUMBER
    )

    test_alignments = align.align_optimal(
        *sequences, matrix, (gap_penalty, gap_penalty), term, local, MAX_NUMBER
    )

    assert test_alignments[0].score == ref_alignments[0].score
    assert len(test_alignments) == len(ref_alignments)
    # We can only expect to get the same alignments in the test and
    # reference, if we get all optimal alignments
    if len(test_alignments) < MAX_NUMBER:
        for alignment in test_alignments:
            try:
                assert alignment in ref_alignments
            except:
                print("Test alignment:")
                print(alignment)
                print()
                print("First reference alignment")
                print(ref_alignments[0])
                raise


@pytest.mark.parametrize(
    "local, term, gap_penalty, seq_indices", itertools.product(
        [True, False], [True, False], [-10, (-10,-1)],
        [(i,j) for i in range(10) for j in range(i+1)]
    )
)
def test_align_optimal_symmetry(sequences, local, term, gap_penalty,
                                seq_indices):
    """
    Alignments should be indifferent about which sequence comes first.
    """
    matrix = align.SubstitutionMatrix.std_protein_matrix()
    index1, index2 = seq_indices
    seq1 = sequences[index1]
    seq2 = sequences[index2]
    alignment1 = align.align_optimal(
        seq1, seq2, matrix,
        gap_penalty=gap_penalty, terminal_penalty=term, local=local,
        max_number=1
    )[0]
    # Swap the sequences
    alignment2 = align.align_optimal(
        seq2, seq1, matrix,
        gap_penalty=gap_penalty, terminal_penalty=term, local=local,
        max_number=1
    )[0]
    # Comparing all traces of both alignments to each other
    # would be unfeasible
    # Instead the scores are compared
    assert alignment1.score == alignment2.score


@pytest.mark.parametrize(
    "gap_penalty, term, seq_indices", itertools.product(
        [-10, (-10,-1)], [False, True],
        [(i,j) for i in range(10) for j in range(i+1)]
    )
)
def test_scoring(sequences, gap_penalty, term, seq_indices):
    """
    The result of the :func:`score()` function should be equal to the
    score calculated in :func:`align_optimal()`.
    """
    matrix = align.SubstitutionMatrix.std_protein_matrix()
    index1, index2 = seq_indices
    seq1 = sequences[index1]
    seq2 = sequences[index2]
    alignment = align.align_optimal(
        seq1, seq2, matrix, gap_penalty=gap_penalty, terminal_penalty=term,
        max_number=1
    )[0]
    try:
        assert align.score(alignment, matrix, gap_penalty, term) \
               == alignment.score
    except AssertionError:
        print(alignment)
        raise


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module provides utility functions for creating filters on atom
arrays and atom array stacks.
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann, Tom Müller"
__all__ = ["filter_solvent", "filter_monoatomic_ions", "filter_nucleotides",
           "filter_amino_acids", "filter_backbone", "filter_intersection",
           "filter_first_altloc", "filter_highest_occupancy_altloc"]

import numpy as np
from .atoms import Atom, AtomArray, AtomArrayStack
from .residues import get_residue_starts
from .info.nucleotides import nucleotide_names


_ext_aa_list = ["ALA","ARG","ASN","ASP","CYS","GLN","GLU","GLY","HIS","ILE",
                "LEU","LYS","MET","PHE","PRO","SER","THR","TRP","TYR","VAL",
                "MSE", "ASX", "GLX", "SEC", "UNK"]

_nucleotide_list = nucleotide_names()

_solvent_list = ["HOH","SOL"]


def filter_monoatomic_ions(array):
    """
    Filter all atoms of an atom array, that are monoatomic ions
    (e.g. sodium or chloride ions).

    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The array to be filtered.

    Returns
    -------
    filter : ndarray, dtype=bool
        This array is `True` for all indices in `array`, where the atom
        is a monoatomic ion.
    """
    # Exclusively in monoatomic ions,
    # the element name is equal to the residue name
    return (array.res_name == array.element)


def filter_solvent(array):
    """
    Filter all atoms of one array that are part of the solvent.

    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The array to be filtered.

    Returns
    -------
    filter : ndarray, dtype=bool
        This array is `True` for all indices in `array`, where the atom
        belongs to the solvent.
    """
    return np.in1d(array.res_name, _solvent_list)


def filter_nucleotides(array):
    """
    Filter all atoms of one array that belong to nucleotides.

    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The array to be filtered.

    Returns
    -------
    filter : ndarray, dtype=bool
        This array is `True` for all indices in `array`, where the atom
        belongs to a nucleotide.
    """
    return np.isin(array.res_name, _nucleotide_list)


def filter_amino_acids(array):
    """
    Filter all atoms of one array that belong to amino acid residues.

    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The array to be filtered.

    Returns
    -------
    filter : ndarray, dtype=bool
        This array is `True` for all indices in `array`, where the atom
        belongs to an amino acid residue.
    """
    return ( np.in1d(array.res_name, _ext_aa_list) & (array.res_id != -1) )


def filter_backbone(array):
    """
    Filter all peptide backbone atoms of one array.

    This includes the "N", "CA" and "C" atoms of amino acids.

    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The array to be filtered.

    Returns
    -------
    filter : ndarray, dtype=bool
        This array is `True` for all indices in `array`, where the atom
        as an backbone atom.
    """
    return ( ((array.atom_name == "N") |
              (array.atom_name == "CA") |
              (array.atom_name == "C")) &
              filter_amino_acids(array) )


def filter_intersection(array, intersect):
    """
    Filter all atoms of one array that exist also in another array.

    An atom is defined as existent in the second array, if there is an
    atom in the second array that has the same annotation values in all
    categories that exists in both arrays.

    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The array to be filtered.
    intersect : AtomArray
        Atoms in `array` that also exists in `intersect` are filtered.

    Returns
    -------
    filter : ndarray, dtype=bool
        This array is `True` for all indices in `array`, where the atom
        exists also in `intersect`.

    Examples
    --------

    Creating an atom array from atoms:

    >>> array1 = AtomArray(length=5)
    >>> array1.chain_id = np.array(["A","B","C","D","E"])
    >>> array2 = AtomArray(length=3)
    >>> array2.chain_id = np.array(["D","B","C"])
    >>> array1 = array1[filter_intersection(array1, array2)]
    >>> print(array1.chain_id)
    ['B' 'C' 'D']

    """
    filter = np.full(array.array_length(), True, dtype=bool)
    intersect_categories = intersect.get_annotation_categories()
    # Check atom equality only for categories,
    # which exist in both arrays
    categories = [category for category in array.get_annotation_categories()
                  if category in intersect_categories]
    for i in range(array.array_length()):
        subfilter = np.full(intersect.array_length(), True, dtype=bool)
        for category in categories:
            subfilter &= (intersect.get_annotation(category)
                          == array.get_annotation(category)[i])
        filter[i] = subfilter.any()
    return filter


def filter_first_altloc(atoms, altloc_ids):
    """
    Filter all atoms, that have the first *altloc* ID appearing in a
    residue.

    Structure files (PDB, PDBx, MMTF) allow for duplicate atom records,
    in case a residue is found in multiple alternate locations
    (*altloc*).
    This function is used to remove such duplicate atoms by choosing a
    single *altloc ID* for an atom with other *altlocs* being removed.

    Parameters
    ----------
    atoms : AtomArray, shape=(n,) or AtomArrayStack, shape=(m,n)
        The unfiltered structure to be filtered.
    altloc_ids : ndarray, shape=(n,), dtype='U1'
        An array containing the alternate location IDs for each
        atom in `atoms`.
        Can contain `'.'`, `'?'`, `' '`, `''` or a letter at each
        position.

    Returns
    -------
    filter : ndarray, dtype=bool
        For each residue, this array is True in the following cases:

            - The atom has no altloc ID (`'.'`, `'?'`, `' '`, `''`).
            - The atom has the same altloc ID (e.g. `'A'`, `'B'`, etc.)
              as the first atom in the residue that has an altloc ID.

    Notes
    -----
    The function will be rarely used by the end user, since this kind
    of filtering is usually automatically performed, when the structure
    is loaded from a file.
    The exception are structures that were read with `altloc` set to
    `True`.

    Examples
    --------

    >>> atoms = array([
    ...     Atom(coord=[1, 2, 3], res_id=1, atom_name="CA"),
    ...     Atom(coord=[4, 5, 6], res_id=1, atom_name="CB"),
    ...     Atom(coord=[6, 5, 4], res_id=1, atom_name="CB")
    ... ])
    >>> altloc_ids = np.array([".", "A", "B"])
    >>> filtered = atoms[filter_first_altloc(atoms, altloc_ids)]
    >>> print(filtered)
                1      CA               1.000    2.000    3.000
                1      CB               4.000    5.000    6.000
    """
    # Filter all atoms without altloc code
    altloc_filter = np.in1d(altloc_ids, [".", "?", " ", ""])

    # And filter all atoms for each residue with the first altloc ID
    residue_starts = get_residue_starts(atoms, add_exclusive_stop=True)
    for start, stop in zip(residue_starts[:-1], residue_starts[1:]):
        letter_altloc_ids = [l for l in altloc_ids[start:stop] if l.isalpha()]
        if len(letter_altloc_ids) > 0:
            first_id = letter_altloc_ids[0]
            altloc_filter[start:stop] |= (altloc_ids[start:stop] == first_id)
        else:
            # No altloc ID in this residue -> Nothing to do
            pass

    return altloc_filter


def filter_highest_occupancy_altloc(atoms, altloc_ids, occupancies):
    """
    For each residue, filter all atoms, that have the *altloc* ID
    with the highest occupancy for this residue.

    Structure files (PDB, PDBx, MMTF) allow for duplicate atom records,
    in case a residue is found in multiple alternate locations
    (*altloc*).
    This function is used to remove such duplicate atoms by choosing a
    single *altloc ID* for an atom with other *altlocs* being removed.

    Parameters
    ----------
    atoms : AtomArray, shape=(n,) or AtomArrayStack, shape=(m,n)
        The unfiltered structure to be filtered.
    altloc_ids : ndarray, shape=(n,), dtype='U1'
        An array containing the alternate location IDs for each
        atom in `atoms`.
        Can contain `'.'`, `'?'`, `' '`, `''` or a letter at each
        position.
    occupancies : ndarray, shape=(n,), dtype=float
        An array containing the occupancy values for each atom in
        `atoms`.

    Returns
    -------
    filter : ndarray, dtype=bool
        For each residue, this array is True in the following cases:

            - The atom has no altloc ID
              (``'.'``, ``'?'``, ``' '``, ``''``).
            - The atom has the altloc ID (e.g. ``'A'``, ``'B'``, etc.),
              of which the corresponding occupancy values are highest
              for the **entire** residue.

    Notes
    -----
    The function will be rarely used by the end user, since this kind
    of filtering is usually automatically performed, when the structure
    is loaded from a file.
    The exception are structures that were read with ``altloc`` set to
    ``True``.

    Examples
    --------

    >>> atoms = array([
    ...     Atom(coord=[1, 2, 3], res_id=1, atom_name="CA"),
    ...     Atom(coord=[4, 5, 6], res_id=1, atom_name="CB"),
    ...     Atom(coord=[6, 5, 4], res_id=1, atom_name="CB")
    ... ])
    >>> altloc_ids = np.array([".", "A", "B"])
    >>> occupancies = np.array([1.0, 0.1, 0.9])
    >>> filtered = atoms[filter_highest_occupancy_altloc(
    ...     atoms, altloc_ids, occupancies
    ... )]
    >>> print(filtered)
                1      CA               1.000    2.000    3.000
                1      CB               6.000    5.000    4.000
    """
    # Filter all atoms without altloc code
    altloc_filter = np.in1d(altloc_ids, [".", "?", " ", ""])

    # And filter all atoms for each residue with the highest sum of
    # occupancies
    residue_starts = get_residue_starts(atoms, add_exclusive_stop=True)
    for start, stop in zip(residue_starts[:-1], residue_starts[1:]):
        occupancies_in_res = occupancies[start:stop]
        altloc_ids_in_res = altloc_ids[start:stop]

        letter_altloc_ids = [l for l in altloc_ids_in_res if l.isalpha()]

        if len(letter_altloc_ids) > 0:
            highest = -1.0
            highest_id = None
            for id in set(letter_altloc_ids):
                occupancy_sum = np.sum(
                    occupancies_in_res[altloc_ids_in_res == id]
                )
                if occupancy_sum > highest:
                    highest = occupancy_sum
                    highest_id = id
            altloc_filter[start:stop] |= (altloc_ids[start:stop] == highest_id)
        else:
            # No altloc ID in this residue -> Nothing to do
            pass

    return altloc_filter



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import pytest
import numpy as np

def pytest_sessionstart(session):
    """
    Compile Cython source files, if Cython is installed and files are
    not compiled, yet.
    """
    try:
        import numpy as np
        import pyximport
        pyximport.install(
            build_in_temp=False,
            setup_args={"include_dirs":np.get_include()},
            language_level=3
        )
    except ImportError:
        pass


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module provides utility for handling data on chain level, rather than
atom level.
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann"
__all__ = ["get_chain_starts", "get_chains", "get_chain_count", "chain_iter"]

import numpy as np


def get_chain_starts(array, add_exclusive_stop=False):
    """
    Get the indices in an atom array, which indicates the beginning of
    a new chain.
    
    A new chain starts, when the chain ID changes.

    This method is internally used by all other chain-related
    functions.
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The atom array (stack) to get the chain starts from.
    add_exclusive_stop : bool, optional
        If true, the exclusive stop of the input atom array, i.e.
        ``array.array_length()``, is added to the returned array of
        start indices as last element.
        
    Returns
    -------
    starts : ndarray, dtype=int
        The start indices of new chains in `array`.
    
    See also
    --------
    get_residue_starts
    """
    # This mask is 'true' at indices where the value changes
    chain_id_changes = (array.chain_id[1:] != array.chain_id[:-1])
    
    # Convert mask to indices
    # Add 1, to shift the indices from the end of a chain
    # to the start of a new chain
    chain_starts = np.where(chain_id_changes)[0] +1
    
    # The first chain is not included yet -> Insert '[0]'
    if add_exclusive_stop:
        return np.concatenate(([0], chain_starts, [array.array_length()]))
    else:
        return np.concatenate(([0], chain_starts))


def get_chains(array):
    """
    Get the chain IDs of an atom array (stack).
    
    The chains are listed in the same order they occur in the array
    (stack).
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The atom array (stack), where the chains are determined.
        
    Returns
    -------
    ids : ndarray, dtype=str
        List of chain IDs.
        
    See also
    --------
    get_residues
    """
    return array.chain_id[get_chain_starts(array)]


def get_chain_count(array):
    """
    Get the amount of chains in an atom array (stack).
    
    The count is determined from the `chain_id` annotation.
    Each time the chain ID changes, the count is incremented.
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The atom array (stack), where the chains are counted.
        
    Returns
    -------
    count : int
        Amount of chains.
    
    See also
    --------
    get_residue_count
    """
    return len(get_chain_starts(array))


def chain_iter(array):
    """
    Iterate over all chains in an atom array (stack).
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The atom array (stack) to iterate over.
        
    Yields
    ------
    chain : AtomArray or AtomArrayStack
        A single chain of the input `array`.
    
    See also
    --------
    residue_iter
    """
    # The exclusive stop is appended to the chain starts
    starts = np.append(get_chain_starts(array), [array.array_length()])
    for i in range(len(starts)-1):
        yield array[..., starts[i] : starts[i+1]]


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from os.path import join
import numpy as np
import pytest
import biotite.structure as struc
import biotite.structure.info as info
import biotite.structure.io as strucio
import biotite.structure.io.mmtf as mmtf
from ..util import data_dir


def generate_random_bond_list(atom_count, bond_count, seed=0):
    """
    Generate a random :class:`BondList`.
    """
    np.random.seed(seed)
    # Create random bonds between atoms of
    # a potential atom array of length ATOM_COUNT
    bonds = np.random.randint(atom_count, size=(bond_count, 3))
    # Clip bond types to allowed BondType values
    bonds[:, 2] %= len(struc.BondType)
    # Remove bonds of atoms to itself
    bonds = bonds[bonds[:,0] != bonds[:,1]]
    assert len(bonds) > 0
    return struc.BondList(atom_count, bonds)


@pytest.fixture(
    params=[False, True] # as_negative
)
def bond_list(request):
    """
    A toy :class:`BondList`.
    """
    as_negative = request.param
    bond_array = np.array([(0,1),(2,1),(3,1),(3,4),(3,1),(1,2),(4,0),(6,4)])
    if as_negative:
        return struc.BondList(7, -7 + bond_array)
    else:
        return struc.BondList(7, bond_array)


def test_creation(bond_list):
    """
    Test creating a :class:`BondList` on a known example.
    """
    # Test includes redundancy removal and max bonds calculation
    assert bond_list.as_array().tolist() == [[0, 1, 0],
                                             [1, 2, 0],
                                             [1, 3, 0],
                                             [3, 4, 0],
                                             [0, 4, 0],
                                             [4, 6, 0]]
    assert bond_list._max_bonds_per_atom == 3
    assert bond_list._atom_count == 7


def test_invalid_creation():
    """
    Check for appropriate errors when invalid input is given to the
    :class:`BondLst` constructor.
    """
    # Test invalid input shapes
    with pytest.raises(ValueError):
        struc.BondList(
            5,
            np.array([
                [1,2,3,4]
            ])
        )
    with pytest.raises(ValueError):
        struc.BondList(
            5,
            np.array([1,2])
        )

    # Test invalid atom indices
    with pytest.raises(IndexError):
        struc.BondList(
            5,
            np.array([
                [1,2],
                # 5 is an invalid index for an atom count of 5
                [5,2]
            ])
        )
    with pytest.raises(IndexError):
        struc.BondList(
            5,
            np.array([
                # Index -6 is invalid for an atom count of 5
                [-6,3],
                [3,4]
            ])
        )
    
    # Test invalid BondType
    with pytest.raises(ValueError):
        struc.BondList(
            5,
            np.array([
                # BondType '8' does not exist
                [1,2,8]
            ])
        )


def test_modification(bond_list):
    """
    Test whether `BondList` correctly identifies whether it contains a
    certain bond on a known example.
    """
    # Already in list
    bond_list.add_bond(3, 1)
    # Also already in list -> update
    bond_list.add_bond(1, 3, 1)
    # The same but with negative atom index
    bond_list.add_bond(-6, -4, 1)
    # Not in list
    bond_list.add_bond(4, 1)
    # In list -> remove
    bond_list.remove_bond(4, 0)
    # Not in list -> Do nothing
    bond_list.remove_bond(0, 3)
    # Remove mutliple bonds, one of them is not in list
    bond_list.remove_bonds(struc.BondList(10, np.array([(1,0),(1,2),(8,9)])))
    assert bond_list.as_array().tolist() == [[1, 3, 1],
                                             [3, 4, 0],
                                             [4, 6, 0],
                                             [1, 4, 0]]


def test_contains(bond_list):
    """
    Test whether `BondList` correctly identifies whether it contains a
    certain bond on a known example.
    """
    for atom1, atom2, _ in bond_list.as_array():
        assert (atom1, atom2) in bond_list
        assert (atom2, atom1) in bond_list
    # Negative test
    assert (0, 5) not in bond_list


def test_access(bond_list):
    """
    Test getting bonds from a `BondList` objects on a known example.
    """
    # Bigger challenge with different bond types
    bond_list.add_bond(1, 3, 1)
    bonds, bond_types = bond_list.get_bonds(0)
    assert bonds.tolist() == [1, 4]
    assert bond_types.tolist() == [0, 0]
    bonds, bond_types = bond_list.get_bonds(1)
    assert bonds.tolist() == [0, 2, 3]
    assert bond_types.tolist() == [0, 0, 1]
    bonds, bond_types = bond_list.get_bonds(2)
    assert bonds.tolist() == [1]
    assert bond_types.tolist() == [0]
    bonds, bond_types = bond_list.get_bonds(3)
    assert bonds.tolist() == [1, 4]
    assert bond_types.tolist() == [1, 0]
    bonds, bond_types = bond_list.get_bonds(4)
    assert bonds.tolist() == [3, 0, 6]
    assert bond_types.tolist() == [0, 0, 0]


def test_merge(bond_list):
    """
    Test merging two `BondList` objects on a known example.
    """
    merged_list = bond_list.merge(struc.BondList(8, np.array([(4,6),(6,7)])))
    assert merged_list.as_array().tolist() == [[0, 1, 0],
                                               [1, 2, 0],
                                               [1, 3, 0],
                                               [3, 4, 0],
                                               [0, 4, 0],
                                               [4, 6, 0],
                                               [6, 7, 0]]


def test_concatenation(bond_list):
    """
    Test concatenation of two `BondList` objects on a known example.
    """
    bond_list += struc.BondList(3, np.array([(0,1,2),(1,2,2)]))
    assert bond_list.as_array().tolist() == [[0, 1, 0],
                                             [1, 2, 0],
                                             [1, 3, 0],
                                             [3, 4, 0],
                                             [0, 4, 0],
                                             [4, 6, 0],
                                             [7, 8, 2],
                                             [8, 9, 2]]
    assert bond_list._max_bonds_per_atom == 3
    assert bond_list._atom_count == 10


def test_indexing(bond_list):
    """
    Test indexing with different index types on a known example.
    """
    sub_list = bond_list[:]
    assert sub_list.as_array().tolist() == bond_list.as_array().tolist()
    sub_list = bond_list[1:6:2]
    assert sub_list.as_array().tolist() == [[0, 1, 0]]
    sub_list = bond_list[:4]
    assert sub_list.as_array().tolist() == [[0, 1, 0],
                                            [1, 2, 0],
                                            [1, 3, 0]]
    sub_list = bond_list[2:]
    assert sub_list.as_array().tolist() == [[1, 2, 0],
                                            [2, 4, 0]]
    
    sub_list = bond_list[[0,3,4]]
    assert sub_list.as_array().tolist() == [[1, 2, 0],
                                            [0, 2, 0]]

    sub_list = bond_list[np.array([True,False,False,True,True,False,True])]
    assert sub_list.as_array().tolist() == [[1, 2, 0],
                                            [0, 2, 0],
                                            [2, 3, 0]]


def test_get_all_bonds():
    """
    Test whether the individual rows returned from
    :func:`get_all_bonds()` are equal to corresponding calls of
    :func:`get_bonds()`.
    """
    ATOM_COUNT = 100
    BOND_COUNT = 500
    
    bond_list = generate_random_bond_list(ATOM_COUNT, BOND_COUNT)

    bonds, bond_types = bond_list.get_all_bonds()

    assert (bonds != -1).all(axis=1).any(axis=0)
    assert (bond_types != -1).all(axis=1).any(axis=0)

    test_bonds = [
        (
            bonded_i[bonded_i != -1].tolist(),
            bond_type[bond_type != -1].tolist()
        )
        for bonded_i, bond_type in zip(bonds, bond_types)
    ]

    ref_bonds = [bond_list.get_bonds(i) for i in range(ATOM_COUNT)]
    ref_bonds = [
        (bonded_i.tolist(), bond_type.tolist())
        for bonded_i, bond_type in ref_bonds
    ]

    assert test_bonds == ref_bonds


def test_adjacency_matrix():
    """
    Test whether the matrix created from :func:`adjacency_matrix()`
    is equal to a manually created matrix using :func:`get_bonds()`.
    """
    ATOM_COUNT = 100
    BOND_COUNT = 500
    
    bond_list = generate_random_bond_list(ATOM_COUNT, BOND_COUNT)

    test_matrix = bond_list.adjacency_matrix()

    ref_matrix = np.zeros((ATOM_COUNT, ATOM_COUNT), dtype=bool)
    for i in range(ATOM_COUNT):
        for j, _ in zip(*bond_list.get_bonds(i)):
            ref_matrix[i, j] = True
    
    assert test_matrix.tolist() == ref_matrix.tolist()


def test_bond_type_matrix():
    """
    Test whether the matrix created from :func:`bond_type_matrix()`
    is equal to a manually created matrix using :func:`get_bonds()`.
    """
    ATOM_COUNT = 100
    BOND_COUNT = 500
    
    bond_list = generate_random_bond_list(ATOM_COUNT, BOND_COUNT)

    test_matrix = bond_list.bond_type_matrix()

    ref_matrix = np.full((ATOM_COUNT, ATOM_COUNT), -1, dtype=int)
    for i in range(ATOM_COUNT):
        for j, bond_type in zip(*bond_list.get_bonds(i)):
            ref_matrix[i, j] = bond_type
    
    assert test_matrix.tolist() == ref_matrix.tolist()


def test_sorted_array_indexing():
    """
    Test whether indexing with an sorted index array results in the
    same `BondList` as indexing with an equivalent boolean mask.
    """
    ATOM_COUNT = 100
    BOND_COUNT = 500
    INDEX_SIZE = 80

    bonds = generate_random_bond_list(ATOM_COUNT, BOND_COUNT)

    # Create a sorted array of random indices for the BondList
    # Indices may not occur multiple times -> 'replace=False'
    index_array = np.sort(np.random.choice(
        np.arange(ATOM_COUNT), INDEX_SIZE, replace=False
    ))
    test_bonds = bonds[index_array]

    # Create a boolean mask that indexes the same elements as the array 
    mask = np.zeros(ATOM_COUNT, dtype=bool)
    mask[index_array] = True
    ref_bonds = bonds[mask]

    assert test_bonds.as_array().tolist() == ref_bonds.as_array().tolist()


def test_unsorted_array_indexing():
    """
    Test whether indexing with an unsorted index array results in the
    same bonded atoms, as indexing with a sorted index array.
    The `BondList` objects should be different because the actual atom
    order is different, but the pointed atoms should still be the same,
    if the respective index is also applied to the atoms.
    """
    ATOM_COUNT = 100
    BOND_COUNT = 500
    INDEX_SIZE = 80

    bonds = generate_random_bond_list(ATOM_COUNT, BOND_COUNT)
    # For simplicity use a reference integer array
    # instead of an atom array
    integers = np.arange(ATOM_COUNT)
    
    # Create random bonds between the reference integers
    bonds = np.random.randint(ATOM_COUNT, size=(BOND_COUNT, 2))
    # Remove bonds of elements to itself
    bonds = bonds[bonds[:,0] != bonds[:,1]]
    assert len(bonds) > 0
    bonds = struc.BondList(ATOM_COUNT, bonds)

    # Create an unsorted array of random indices for the BondList
    # Indices should be unsorted -> 'replace=False'
    unsorted_index = np.random.choice(
        np.arange(ATOM_COUNT), INDEX_SIZE, replace=False
    )
    test_bonds = bonds[unsorted_index]

    # Create a sorted variant of the index array
    sorted_index = np.sort(unsorted_index)
    # Check whether the unsorted array is really unsorted
    assert sorted_index.tolist() != unsorted_index.tolist()
    ref_bonds = bonds[sorted_index]

    unsorted_indexed_integers = integers[unsorted_index]
    sorted_indexed_integers = integers[sorted_index]
    # Get the 'atoms', in this case integers, that are connected with a bond
    # Use a set for simpler comparison between the sorted and unsorted variant
    # Omit the bond type -> 'bonds.as_array()[:, :2]'
    test_integer_pairs = set([
        frozenset((unsorted_indexed_integers[i], unsorted_indexed_integers[j]))
        for i, j in test_bonds.as_array()[:, :2]
    ])
    ref_integer_pairs = set([
        frozenset((sorted_indexed_integers[i], sorted_indexed_integers[j]))
        for i, j in ref_bonds.as_array()[:, :2]
    ])

    # The BondList entries should be different,
    # since they point to different positions in the reference array
    assert test_bonds.as_array().tolist() != ref_bonds.as_array().tolist()
    # But the actual bonded 'atom' pairs, should still be the same
    assert test_integer_pairs == ref_integer_pairs
    # Additionally, check whether in each bond the lower atom index 
    # comes first
    for i, j, _ in test_bonds.as_array():
        assert i < j


def test_atom_array_consistency():
    """
    Test whether the associated bonds of an `AtomArray` still point to
    the same atoms after indexing with a boolean mask.
    The boolean mask is constructed in a way that all bonded atoms are
    masked.
    """
    array = strucio.load_structure(join(data_dir("structure"), "1l2y.mmtf"))[0]
    ca = array[array.atom_name == "CA"]
    # Just for testing, does not reflect real bonds
    bond_list = struc.BondList(ca.array_length(), 
        np.array([(0,1),(2,8),(5,15),(1,5),(0,9),(3,18),(2,9)])
    )
    ca.bonds = bond_list
    
    ref_ids = ca.res_id[bond_list.as_array()[:,:2].flatten()]
    
    # Some random boolean mask as index,
    # but all bonded atoms are included
    mask = np.array([1,1,1,1,0,1,0,0,1,1,0,1,1,0,0,1,1,0,1,1], dtype=bool)
    masked_ca = ca[mask]
    test_ids = masked_ca.res_id[masked_ca.bonds.as_array()[:,:2].flatten()]
    
    # The bonds, should always point to the same atoms (same res_id),
    # irrespective of indexing
    assert test_ids.tolist() == ref_ids.tolist()


@pytest.mark.parametrize("single_model", [False, True])
def test_connect_via_residue_names(single_model):
    """
    Test whether the created bond list is equal to the bonds deposited
    in the MMTF file.
    """
    # Structure with peptide, nucleotide, small molecules and water
    file = mmtf.MMTFFile.read(join(data_dir("structure"), "5ugo.mmtf"))
    if single_model:
        atoms = mmtf.get_structure(file, include_bonds=True, model=1)
    else:
        atoms = mmtf.get_structure(file, include_bonds=True)
    
    ref_bonds = atoms.bonds

    test_bonds = struc.connect_via_residue_names(atoms)
    # MMTF format does not represent aromaticity
    test_bonds.remove_aromaticity()

    assert test_bonds == ref_bonds


def test_connect_via_distances():
    """
    Test whether the created bond list is equal to the bonds deposited
    in the MMTF file.
    """
    file = mmtf.MMTFFile.read(join(data_dir("structure"), "1l2y.mmtf"))
    atoms = mmtf.get_structure(file, include_bonds=True, model=1)
    # Remove termini to solve the issue that the reference bonds do not
    # contain proper bonds for the protonated/deprotonated termini
    atoms = atoms[(atoms.res_id > 1) & (atoms.res_id < 20)]
    
    ref_bonds = atoms.bonds
    # Convert all bonds to BondType.ANY
    ref_bonds = struc.BondList(
        ref_bonds.get_atom_count(), ref_bonds.as_array()[:, :2]
    )

    test_bonds = struc.connect_via_distances(atoms)

    assert test_bonds == ref_bonds


def test_find_connected(bond_list):
    """
    Find all connected atoms to an atom in a known example.
    """
    for index in (0,1,2,3,4,6):
        assert struc.find_connected(bond_list, index).tolist() == [0,1,2,3,4,6]
    assert struc.find_connected(bond_list, 5).tolist() == [5]


@pytest.mark.parametrize(
    "res_name, expected_bonds",
    [
        # Easy ligand visualization at:
        # https://www.rcsb.org/ligand/<ABC>
        ("TYR", [
            ("N",   "CA" ),
            ("CA",  "C"  ),
            ("CA",  "CB" ),
            ("C",   "OXT"),
            ("CB",  "CG" ),
            ("CZ",  "OH" ),
        ]),
        ("CEL", [
            ("C1",   "C4" ),
            ("C8",   "C11"),
            ("C15",  "S1" ),
            ("N3",   "S1" )
        ]),
        ("LEO", [
            ("C3",   "C8" ),
            ("C6",   "C17"),
            ("C17",  "C22"),
        ]),
    ]
)
def test_find_rotatable_bonds(res_name, expected_bonds):
    """
    Check the :func:`find_rotatable_bonds()` function based on
    known examples.
    """
    molecule = info.residue(res_name)
    
    ref_bond_set = {
        tuple(sorted((name_i, name_j))) for name_i, name_j in expected_bonds
    }

    rotatable_bonds = struc.find_rotatable_bonds(molecule.bonds)
    test_bond_set = set()
    for i, j, _ in rotatable_bonds.as_array():
        test_bond_set.add(
            tuple(sorted((molecule.atom_name[i], molecule.atom_name[j])))
        )
    
    # Compare with reference bonded atom names
    assert test_bond_set == ref_bond_set
    # All rotatable bonds must be single bonds
    assert np.all(rotatable_bonds.as_array()[:, 2] == struc.BondType.SINGLE)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import itertools
import glob
import io
import biotite.sequence as seq
import biotite.sequence.io.fasta as fasta
import numpy as np
import os
import os.path
from ..util import data_dir
import pytest

   
def test_access_low_level():
    path = os.path.join(data_dir("sequence"), "nuc.fasta")
    file = fasta.FastaFile.read(path)
    assert file["dna sequence"] == "ACGCTACGT"
    assert file["another dna sequence"] == "A"
    assert file["third dna sequence"] == "ACGT"
    assert dict(file.items()) == {
        "dna sequence" : "ACGCTACGT",
        "another dna sequence" : "A",
        "third dna sequence" : "ACGT",
        "rna sequence" : "ACGU",
        "ambiguous rna sequence" : "ACGUNN",
    }
    file["another dna sequence"] = "AA"
    del file["dna sequence"]
    file["yet another sequence"] = "ACGT"
    assert dict(file.items()) == {
        "another dna sequence" : "AA",
        "third dna sequence" : "ACGT",
        "rna sequence" : "ACGU",
        "ambiguous rna sequence" : "ACGUNN",
        "yet another sequence" : "ACGT",
    }


def test_access_high_level():
    path = os.path.join(data_dir("sequence"), "nuc.fasta")
    file = fasta.FastaFile.read(path)
    sequences = fasta.get_sequences(file)
    assert sequences == {
        "dna sequence" : seq.NucleotideSequence("ACGCTACGT", False),
        "another dna sequence" : seq.NucleotideSequence("A", False),
        "third dna sequence" : seq.NucleotideSequence("ACGT", False),
        "rna sequence" : seq.NucleotideSequence("ACGT", False),
        "ambiguous rna sequence" : seq.NucleotideSequence("ACGTNN", True),
    }


def test_sequence_conversion():
    path = os.path.join(data_dir("sequence"), "nuc.fasta")
    file = fasta.FastaFile.read(path)
    assert seq.NucleotideSequence("ACGCTACGT") == fasta.get_sequence(file)
    
    seq_dict = fasta.get_sequences(file)
    file2 = fasta.FastaFile()
    fasta.set_sequences(file2, seq_dict)
    seq_dict2 = fasta.get_sequences(file2)
    # Cannot compare dicts directly, since the original RNA sequence is
    # now guessed as protein sequence
    for seq1, seq2 in zip(seq_dict.values(), seq_dict2.values()):
        assert str(seq1) == str(seq2)
    
    file3 = fasta.FastaFile()
    fasta.set_sequence(file3, seq.NucleotideSequence("AACCTTGG"))
    assert file3["sequence"] == "AACCTTGG"
    
    path = os.path.join(data_dir("sequence"), "prot.fasta")
    file4 = fasta.FastaFile.read(path)
    # Expect a warning for selenocysteine conversion
    with pytest.warns(UserWarning):
        assert seq.ProteinSequence("YAHCGFRTGS") == fasta.get_sequence(file4)
    
    path = os.path.join(data_dir("sequence"), "invalid.fasta")
    file5 = fasta.FastaFile.read(path)
    with pytest.raises(ValueError):
        seq.NucleotideSequence(fasta.get_sequence(file5))


def test_rna_conversion():
    sequence = seq.NucleotideSequence("ACGT")
    fasta_file = fasta.FastaFile()
    fasta.set_sequence(fasta_file, sequence, "seq1", as_rna=False)
    fasta.set_sequence(fasta_file, sequence, "seq2", as_rna=True)
    assert fasta_file["seq1"] == "ACGT"
    assert fasta_file["seq2"] == "ACGU"


def test_alignment_conversion():
    path = os.path.join(data_dir("sequence"), "alignment.fasta")
    file = fasta.FastaFile.read(path)
    alignment = fasta.get_alignment(file)
    assert str(alignment) == ("ADTRCGTARDCGTR-DRTCGRAGD\n"
                              "ADTRCGT---CGTRADRTCGRAGD\n"
                              "ADTRCGTARDCGTRADR--GRAGD")
    
    file2 = fasta.FastaFile()
    fasta.set_alignment(file2, alignment, seq_names=["seq1","seq2","seq3"])
    alignment2 = fasta.get_alignment(file2)
    assert str(alignment) == str(alignment2)

@pytest.mark.parametrize(
    "file_name",
    glob.glob(os.path.join(data_dir("sequence"), "*.fasta"))
)
def test_read_iter(file_name):
    ref_dict = dict(fasta.FastaFile.read(file_name).items())
    
    test_dict = dict(fasta.FastaFile.read_iter(file_name))

    assert test_dict == ref_dict


@pytest.mark.parametrize(
    "chars_per_line, n_sequences", itertools.product(
        [80, 200],
        [1, 10]
    )
)
def test_write_iter(chars_per_line, n_sequences):
    """
    Test whether :class:`FastaFile.write()` and
    :class:`FastaFile.write_iter()` produce the same output file for
    random sequences.
    """
    LENGTH_RANGE = (50, 150)
    SCORE_RANGE = (10, 60)

    # Generate random sequences and scores
    np.random.seed(0)
    sequences = []
    for i in range(n_sequences):
        seq_length = np.random.randint(*LENGTH_RANGE)
        code = np.random.randint(
            len(seq.NucleotideSequence.alphabet_unamb),
            size=seq_length
        )
        sequence = seq.NucleotideSequence()
        sequence.code = code
        sequences.append(sequence)
    
    fasta_file = fasta.FastaFile(chars_per_line)
    for i, sequence in enumerate(sequences):
        header = f"seq_{i}"
        fasta_file[header] = str(sequence)
    ref_file = io.StringIO()
    fasta_file.write(ref_file)
    
    test_file = io.StringIO()
    fasta.FastaFile.write_iter(
        test_file,
        (
            (f"seq_{i}", str(sequence))
            for i, sequence in enumerate(sequences)
        ),
        chars_per_line
    )

    assert test_file.getvalue() == ref_file.getvalue()


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import warnings
from tempfile import TemporaryFile
import glob
from os.path import join
import pytest
import numpy as np
import biotite.structure as struc
import biotite.structure.io.pdbqt as pdbqt
import biotite.structure.io.pdbx as pdbx
from ..util import data_dir


@pytest.mark.parametrize(
    "path", glob.glob(join(data_dir("structure"), "*.cif"))
)
def test_array_conversion(path):
    pdbx_file = pdbx.PDBxFile.read(path)
    ref_structure = pdbx.get_structure(
        pdbx_file, model=1, extra_fields=["charge"]
    )
    ref_structure.bonds = struc.connect_via_residue_names(ref_structure)

    pdbqt_file = pdbqt.PDBQTFile()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # Ignore warnings about atoms not parametrized 
        mask = pdbqt.set_structure(pdbqt_file, ref_structure)
    ref_structure = ref_structure[mask]
    temp = TemporaryFile("r+")
    pdbqt_file.write(temp)

    temp.seek(0)
    pdbqt_file = pdbqt.PDBQTFile.read(temp)
    test_structure = pdbqt.get_structure(pdbqt_file, model=1)
    temp.close()

    assert np.allclose(test_structure.coord, ref_structure.coord)
    for category in test_structure.get_annotation_categories():
        if category == "element":
            # PDBQT uses special atom types, which replace the usual
            # elements
            # -> there cannot be equality of the 'element' annotation
            continue
        try:
            assert np.array_equal(
                test_structure.get_annotation(category),
                 ref_structure.get_annotation(category)
            )
        except AssertionError:
            print(f"Inequality in '{category}' category")
            raise



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.info"
__author__ = "Patrick Kunzmann"
__all__ = ["all_residues", "full_name", "link_type"]

from os.path import join, dirname, realpath
import msgpack


_info_dir = dirname(realpath(__file__))
# Data is taken from
# ftp://ftp.wwpdb.org/pub/pdb/data/monomers/components.cif
# (2019/01/27)
with open(join(_info_dir, "residue_names.msgpack"), "rb") as file:
    _res_names = msgpack.load(file, raw=False)
with open(join(_info_dir, "link_types.msgpack"), "rb") as file:
    _link_types = msgpack.load(file, raw=False)


def all_residues():
    """
    Get a list of all residues/compound names in the
    PDB chemical components dictionary.
    
    Returns
    -------
    residues : list of str
        A list of all available The up to 3-letter residue names.
    
    Examples
    --------

    >>> print(all_residues()[1000 : 1010])
    ['0Y4', '0Y5', '0Y7', '0Y8', '0Y9', '0YA', '0YB', '0YC', '0YD', '0YE']
    """
    return list(_res_names.keys()) 


def full_name(res_name):
    """
    Get the full name of a residue/compound from the up to 3-letter
    residue name, based on the PDB chemical components dictionary.

    Parameters
    ----------
    res_name : str
        The up to 3-letter residue name.
    
    Returns
    -------
    name : str
        The full name of the residue.
    
    Examples
    --------

    >>> print(full_name("MAN"))
    alpha-D-mannopyranose
    """
    return _res_names.get(res_name.upper())


def link_type(res_name):
    """
    Get the linking type of a residue/compound,
    based on the PDB chemical components dictionary.

    Parameters
    ----------
    res_name : str
        The up to 3-letter residue name.
    
    Returns
    -------
    link_type : str
        The link type.
    
    Examples
    --------

    >>> print(link_type("MAN"))
    D-saccharide, alpha linking
    >>> print(link_type("TRP"))
    L-PEPTIDE LINKING
    >>> print(link_type("HOH"))
    NON-POLYMER
    """
    return _link_types.get(res_name.upper())


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import biotite.sequence as seq
import numpy as np
import pytest


def test_nucleotide_construction():
    string = "AATGCGTTA"
    string_amb = "ANNGCBRTAN"
    dna = seq.NucleotideSequence(string)
    assert dna.get_alphabet() == seq.NucleotideSequence.alphabet_unamb
    assert str(dna) == string
    dna = seq.NucleotideSequence(string_amb)
    assert dna.get_alphabet() == seq.NucleotideSequence.alphabet_amb
    assert str(dna) == string_amb


def test_reverse_complement():
    string = "AATGCGTTANY"
    dna = seq.NucleotideSequence(string)
    assert str(dna.reverse().complement()) == "RNTAACGCATT"

def test_stop_removal():
    string = "LYG*GR*"
    protein = seq.ProteinSequence(string)
    assert str(protein.remove_stops()) == string.replace("*", "")


@pytest.mark.parametrize("dna_str, protein_str",
                         [("CACATAGCATGA", "HIA*"),
                          ("ATGTAGCTA", "M*L")])
def test_full_translation(dna_str, protein_str):
    dna = seq.NucleotideSequence(dna_str)
    protein = dna.translate(complete=True)
    assert protein_str == str(protein)


@pytest.mark.parametrize("dna_str, protein_str_list",
                         [("CA", []),
                          ("GAATGCACTGAGATGCAATAG", ["MH*","MQ*"]),
                          ("ATGCACATGTAGGG", ["MHM*","M*"]),
                          ("GATGCATGTGAAAA", ["MHVK","M*"])])
def test_frame_translation(dna_str, protein_str_list):
    dna = seq.NucleotideSequence(dna_str)
    proteins, pos = dna.translate(complete=False)
    assert len(proteins) == len(protein_str_list)
    assert set([str(protein) for protein in proteins]) == set(protein_str_list)
    # Test if the positions are also right
    # -> Get sequence slice and translate completely
    assert set([str(dna[start : stop].translate(complete=True))
                for start, stop in pos]
    ) == set(protein_str_list)


def test_translation_met_start():
    """
    Test whether the start amino acid is replaced by methionine,
    i.e. the correct function of the 'met_start' parameter.
    """
    codon_table = seq.CodonTable.default_table().with_start_codons("AAA")
    dna = seq.NucleotideSequence("GAAACTGAAATAAGAAC")
    proteins, _ = dna.translate(codon_table=codon_table, met_start=True)
    assert [str(protein) for protein in proteins] == ["MLK*", "M*"]


def test_letter_conversion():
    for symbol in seq.ProteinSequence.alphabet:
        three_letters = seq.ProteinSequence.convert_letter_1to3(symbol)
        single_letter = seq.ProteinSequence.convert_letter_3to1(three_letters)
        assert symbol == single_letter


@pytest.mark.parametrize(
    "monoisotopic, expected_mol_weight_protein",
    # Reference values taken from https://web.expasy.org/compute_pi/
    [(True, 2231.06), (False, 2232.56)]
)
def test_get_molecular_weight(monoisotopic, expected_mol_weight_protein):
    """
    Test whether the molecular weight of a protein is calculated
    correctly.
    """
    protein = seq.ProteinSequence("ACDEFGHIKLMNPQRSTVW")
    mol_weight_protein = protein.get_molecular_weight(
        monoisotopic=monoisotopic)
    assert mol_weight_protein == \
           pytest.approx(expected_mol_weight_protein, abs=1e-2)




# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.graphics"
__author__ = "Patrick Kunzmann"
__all__ = ["plot_feature_map", "FeaturePlotter", "MiscFeaturePlotter",
           "CodingPlotter", "PromoterPlotter", "TerminatorPlotter",
           "RBSPlotter"]

import copy
import abc
import numpy as np
from ...visualize import colors, AdaptiveFancyArrow
from ..annotation import Annotation, Feature, Location


def plot_feature_map(axes, annotation, loc_range=None,
                     multi_line=True, symbols_per_line=1000,
                     show_numbers=False, number_size=None, line_width=0.05,
                     show_line_position=False, spacing=0.25,
                     feature_plotters=None, style_param=None):
    """
    Plot a sequence annotation, by showing the range of each feature
    on one or multiple position depicting line(s).

    This function uses :class:`FeaturePlotter` objects to draw the
    features.
    This function internally uses a list of plotters, where the
    first plotter in the list, that supports a feature, is used to draw
    that feature.
    The amount of features that can be visualized by default is limited.
    Features, that are not supported otherwise, are visualized as simple
    rectangles.
    Additional :class:`FeaturePlotter` objects, that are supplied in the
    `feature_plotters` parameters, can be used to add support for
    further features or to customize the appearance of certain features.

    Parameters
    ----------
    axes : Axes
        A *Matplotlib* axes, that is used as plotting area.
    annotation : Annotation
        The annotation to be visualized.
    loc_range : tuple (int, int), optional
        The start and exclusive stop location that is visualized.
        By default, the location range starts from the first
        base/residue and ends at the last base/residue of all features,
        ensuring that the entire annotation is drawn.
    multi_line : bool, optional
        If true, the annotation is segmented into multiple lines with a
        line break all `symbols_per_line` lines.
        Otherwise, the entire location range is put into a single line.
    symbols_per_line : int, optional
        The amount of
        Does not have an effect, if `multi_line` is false.
    show_numbers : bool, optional
        If true, the sequence position the base/residue of a line is
        shown on the right side of the plot.
    number_size : float, optional
        The font size of the position numbers
    line_width : float, optional
        The size of the continuous line as fraction of the height of
        the drawn features.
    show_line_position : bool, optional
        If true the position within a line is plotted.
    spacing : float, optional
        The size of the spacing between the lines as fraction of the
        height of the drawn features.
    feature_plotters : list of FeaturePlotter, optional
        Custom plotters for features.
        The list is iterated from the beginning until a
        :class:`FeaturePlotter` matches the respective feature
        (`FeaturePlotter.matches()` returns `True`).
        This :class:`FeaturePlotter` is then used to draw the feature.
        Therefore, the :class:`FeaturePlotter` instances in the list
        have descending priority.
        The default plotters are appended after this supplied list,
        i.e. the default plotters have a lower priority.
    style_param : dict
        Additional style parameters that are given to the
        :class:`FeaturePlotter` objects.

    Notes
    -----
    Good visulation results are obtained only for non-overlapping
    features.
    When two features overlap, their drawing area does also overlap.
    """
    from matplotlib.transforms import Bbox
    from matplotlib.patches import Rectangle

    if loc_range is None:
        loc_range = annotation.get_location_range()
    loc_range_length = loc_range[1] - loc_range[0]
    if multi_line:
        symbols_per_line = symbols_per_line
    else:
        # Line length covers the entire location range
        symbols_per_line = loc_range_length
    
    plotters = [
        PromoterPlotter(),
        TerminatorPlotter(),
        RBSPlotter(),
        CodingPlotter(),
        MiscFeaturePlotter()
    ]
    if feature_plotters is not None:
        plotters = list(feature_plotters) + plotters

    style_param = {} if style_param is None else style_param

    line_count = loc_range_length // symbols_per_line
    # Only extend line count by 1 if there is a remainder
    if loc_range_length % symbols_per_line != 0:
        line_count += 1


    ### Draw lines ###
    remaining_symbols = loc_range_length
    y = 0.5
    while remaining_symbols > 0:
        if remaining_symbols > symbols_per_line:
            # Line spans the entire plot (horizontally)
            line_length = symbols_per_line
        else:
            # Last line -> Line spans to end of annotation
            line_length = remaining_symbols
        axes.add_patch(Rectangle(
            (0, y-line_width/2), line_length, line_width,
            color="gray", linewidth=0
        ))
        # Increment by spacing and width (=1) of feature
        y += spacing + 1
        remaining_symbols -= symbols_per_line
    
    ### Draw features ###
    line_start_loc = loc_range[0]
    y = 0
    while line_start_loc < loc_range[1]:
        annotation_for_line = annotation[
            line_start_loc : line_start_loc + symbols_per_line
        ]
        for feature in annotation_for_line:
            plotter = None
            # Identify fitting plotter
            for potentail_plotter in plotters:
                if potentail_plotter.matches(feature):
                    # Take first fitting plotter in list
                    plotter = potentail_plotter
                    break
            if plotter is not None:
                for loc in feature.locs:
                    loc_len = loc.last - loc.first + 1
                    # Get start location realtive to start if line
                    loc_in_line = loc.first - line_start_loc
                    x = loc_in_line
                    # Line width multiplied by percentage of line
                    width = loc_len
                    height = 1
                    bbox = Bbox.from_bounds(x, y, width, height)
                    plotter.draw(
                        axes, feature, bbox, loc,
                        style_param=style_param
                    )
        # Increment by spacing and width (=1) of feature
        y += spacing + 1
        remaining_symbols += symbols_per_line
        line_start_loc += symbols_per_line
    
    ### Draw position numbers  ###
    ticks = []
    tick_labels = []
    if show_numbers:
        # Numbers at center height of each feature line -> 0.5
        y = 0.5
        for i in range(line_count):
            if i == line_count-1:
                # Last line -> get number of last column in trace
                loc = loc_range[1] -1
            else:
                loc = loc_range[0] + ((i+1) * symbols_per_line) -1
            ticks.append(y)
            tick_labels.append(str(loc))
            # Increment by spacing and width of feature (1)
            y += spacing + 1
    axes.set_yticks(ticks)
    axes.set_yticklabels(tick_labels)


    axes.set_xlim(0, symbols_per_line)
    # Y-axis starts from top
    axes.set_ylim(1*line_count + spacing*(line_count-1), 0)
    axes.set_frame_on(False)
    # Draw location numbers on right side
    axes.get_yaxis().set_tick_params(
        left=False, right=False, labelleft=False, labelright=True
    )
    # Remove ticks and set number font size
    axes.yaxis.set_tick_params(
        left=False, right=False, labelsize=number_size
    )
    
    if show_line_position:
        axes.xaxis.set_tick_params(
            top=False, bottom=True, labeltop=False, labelbottom=True
        )
    else:
        axes.xaxis.set_tick_params(
            top=False, bottom=False, labeltop=False, labelbottom=False
        )


class FeaturePlotter(metaclass=abc.ABCMeta):
    """
    A :class:`FeaturePlotter` is an object, that 'knows' how to draw a
    certain type of sequence feature onto a :class:`matplotlib.Axes`.

    Whether the :class:`FeaturePlotter` is able to draw a feature, is
    checked via the :func:`matches()` method.
    The visualization of a compatible feature is conducted in the
    :func:`draw()` method.
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def matches(self, feature):
        """
        Check, whether this object is able to draw a given sequence
        feature.

        Parameters
        ----------
        feature : Feature
            The sequence feature to be checked.
        
        Returns
        -------
        compatibility : bool
            True, if this object is able to draw the given `feature`,
            false otherwise.
        """
        pass
    
    @abc.abstractmethod
    def draw(self, axes, feature, bbox, location, style_param):
        """
        Draw a feature onto an axes.

        Parameters
        ----------
        axes : Axes
            A *Matplotlib* axes, that is used as plotting area.
        feature : Feature
            The feature to be drawn.
        bbox : Bbox
            The bounding box, that describes the area on the `axes`,
            where the feature should be drawn.
        location : int
            The location of the feature, that should be drawn.
            Might be useful, when the visualization is dependent
            on e.g. location defects.
        style_param : dict
            Additional style parameters.
        """
        pass


class CodingPlotter(FeaturePlotter):
    """
    A plotter for the *CDS* and *gene* features.

    Draws an arrow with a 90 degrees tip.

    Parameters
    ----------
    tail_width : float, optional
        The width of the arrow tail
        as fraction of the feature drawing area height.
    head_width : float, optional
        The width of the arrow head
        as fraction of the feature drawing area height.
    """
    
    def __init__(self, tail_width=0.5, head_width=0.8):
        self._tail_width = tail_width
        self._head_width = head_width

    def matches(self, feature):
        if feature.key in ["CDS", "gene"]:
            return True
        else:
            return False
        
    def draw(self, axes, feature, bbox, loc, style_param):
        y = bbox.y0 + bbox.height/2
        dy = 0
        if loc.strand == Location.Strand.FORWARD:
            x = bbox.x0
            dx = bbox.width
        else:
            x = bbox.x1
            dx = -bbox.width
        
        if  (
                loc.strand == Location.Strand.FORWARD 
                and loc.defect & Location.Defect.MISS_RIGHT
            ) or (
                loc.strand == Location.Strand.REVERSE 
                and loc.defect & Location.Defect.MISS_LEFT
            ):
                # If the feature extends into the prevoius or next line
                # do not draw an arrow head
                draw_head = False
        else:
                draw_head = True
        
        # Create head with 90 degrees tip -> head width/length ratio = 1/2
        axes.add_patch(AdaptiveFancyArrow(
            x, y, dx, dy, self._tail_width, self._head_width, head_ratio=0.5,
            draw_head=draw_head, color=colors["dimgreen"], linewidth=0
        ))

        if feature.key == "CDS":
            if "product" not in feature.qual:
                label = None
            elif feature.qual["product"] == "hypothetical protein":
                label = None
            else:
                label = feature.qual["product"]
        elif feature.key == "gene":
            if  "gene" not in feature.qual:
                label = None
            else:
                label = feature.qual["gene"]
        
        if label is not None:
            center_x = bbox.x0 + bbox.width/2
            center_y = bbox.y0 + bbox.height/2
            axes.text(
                center_x, center_y, label, color="black",
                ha="center", va="center", size=11)


class MiscFeaturePlotter(FeaturePlotter):
    """
    A plotter that matches any feature.

    Draws a simple Rectangle.

    Parameters
    ----------
    height : float, optional
        The width of the rectangle
        as fraction of the feature drawing area height.
    """

    def __init__(self, height=0.4):
        self._height = height

    def matches(self, feature):
        return True
        
    def draw(self, axes, feature, bbox, loc, style_param):
        from matplotlib.patches import Rectangle

        rect = Rectangle(
            (bbox.x0, bbox.y0 + bbox.height/2 * (1-self._height)),
            bbox.width, bbox.height*self._height,
            color=colors["dimorange"], linewidth=0
        )
        axes.add_patch(rect)

class PromoterPlotter(FeaturePlotter):
    """
    A plotter for *regulatory* features with the *promoter* or
    *TATA_box* class.

    Draws a simple curved thin black arrow.

    Parameters
    ----------
    line_width : float, optional
        The width of the curved arrow tail.
    head_width : float, optional
        The width of the arrow head
    head_length : float, optional
        The length of the arrow.
    head_height : float, optional
        The Y-position of the arrow head
        as fraction of the halffeature drawing area height.
    """

    def __init__(self, line_width=2, head_width=2,
                 head_length=6, head_height=0.8):
        self._line_width = line_width
        self._head_width = head_width
        self._head_length = head_length
        self._head_height = head_height

    def matches(self, feature):
        if feature.key == "regulatory":
            if "regulatory_class" in feature.qual:
                if feature.qual["regulatory_class"] in ["promoter","TATA_box"]:
                    return True
        return False
        
    def draw(self, axes, feature, bbox, loc, style_param):
        from matplotlib.patches import FancyArrowPatch, ArrowStyle
        from matplotlib.path import Path

        x_center = bbox.x0 + bbox.width/2
        y_center = bbox.y0 + bbox.height/2

        path = Path(
            vertices=[
                (bbox.x0, y_center),
                (bbox.x0, y_center - bbox.height/2 * self._head_height),
                (bbox.x1, y_center - bbox.height/2 * self._head_height),
            ],
            codes=[
                Path.MOVETO,
                Path.CURVE3,
                Path.CURVE3
            ]
        )
        style = ArrowStyle.CurveFilledB(
            head_width=self._head_width, head_length=self._head_length
        )
        arrow = FancyArrowPatch(
            path=path, arrowstyle=style, linewidth=self._line_width,
            color="black"
        )
        axes.add_patch(arrow)
        
        if "note" in feature.qual:
            axes.text(
                x_center, y_center + bbox.height/4, feature.qual["note"],
                color="black", ha="center", va="center",
                size=9
            )


class TerminatorPlotter(FeaturePlotter):
    """
    A plotter for *regulatory* features with the *terminator* class.

    Draws a vertical bar.

    Parameters
    ----------
    bar_width : float, optional
        The width of the line representing the bar.
    """

    def __init__(self, bar_width=5):
        self._bar_width = bar_width

    def matches(self, feature):
        if feature.key == "regulatory":
            if "regulatory_class" in feature.qual:
                if feature.qual["regulatory_class"] == "terminator":
                    return True
        return False
        
    def draw(self, axes, feature, bbox, loc, style_param):

        x = bbox.x0 + bbox.width/2

        axes.plot(
            (x, x), (bbox.y0, bbox.y1), color="black",
            linestyle="-", linewidth=self._bar_width, marker="None"
        )


class RBSPlotter(FeaturePlotter):
    """
    A plotter for *regulatory* features with the
    *ribosome_binding_site* class.

    Draws an ellipse.

    Parameters
    ----------
    height : float, optional
        The width of the ellipse
        as fraction of the feature drawing area height.
    """

    def __init__(self, height=0.4):
        self._height = height

    def matches(self, feature):
        if feature.key == "regulatory":
            if "regulatory_class" in feature.qual:
                if feature.qual["regulatory_class"] == "ribosome_binding_site":
                    return True
        return False
        
    def draw(self, axes, feature, bbox, loc, style_param):
        from matplotlib.patches import Ellipse

        ellipse = Ellipse(
            (bbox.x0 + bbox.width/2, bbox.y0 + bbox.height/2),
            bbox.width, self._height*bbox.height,
            color=colors["dimorange"], linewidth=0)
        axes.add_patch(ellipse)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application.blast"
__author__ = "Patrick Kunzmann"
__all__ = ["BlastAlignment"]

from ...sequence.align.alignment import Alignment


class BlastAlignment(Alignment):
    """
    A specialized :class:`Alignment` class for alignments using the
    BLAST application. It stores additional data, like the E-value,
    the HSP position and a description of the hit sequence.
    
    Like its superclass, all attributes of a :class:`BlastAlignment` are
    public. The attributes are the same as the constructor parameters.
    
    Parameters
    ----------
    sequences : list
        A list of aligned sequences. Does actually not contain the
        complete original sequences, but the HSP sequences.
    trace : ndarray, dtype=int, shape=(n,m)
        The alignment trace.
    score : int
        Alignment score.
    e_value : float
        Expectation value for the number of random sequences of a
        similar sized database getting an equal or higher score by
        change when aligned with the query sequence.
    query_interval : tuple of int
        Describes the position of the HSP part of the query sequence
        in the original query sequence. The first element is the start
        position, the second element is the inclusive stop position.
        Indexing starts at 1.
    hit_interval : tuple of int
        Analogous to `query_interval`, this describes the position of
        the HSP part of the hit sequence in the complete hit sequence.
    hit_id : str
        The NCBI *unique identifier* (UID) of the hit sequence.
    hit_definition : str
        The name of the hit sequence.
    """
    
    def __init__(self, sequences, trace, score, e_value,
                 query_interval, hit_interval, hit_id, hit_definition):
        super().__init__(sequences, trace, score)
        self.e_value = e_value
        self.query_interval = query_interval
        self.hit_interval = hit_interval
        self.hit_id = hit_id
        self.hit_definition = hit_definition
    
    def __eq__(self, item):
        if not isinstance(item, BlastAlignment):
            return False
        if self.e_value != item.e_value:
            return False
        if self.query_interval != item.query_interval:
            return False
        if self.hit_interval != item.hit_interval:
            return False
        if self.hit_id != item.hit_id:
            return False
        if self.hit_definition != item.hit_definition:
            return False
        return super().__eq__(item)
    
    def __getitem__(self, index):
        super_alignment = super().__getitem__(index)
        return BlastAlignment(
            super_alignment.sequences,
            super_alignment.trace,
            super_alignment.score,
            self.e_value,
            self.query_interval,
            self.hit_interval,
            self.hit_id,
            self.hit_definition
        )


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import itertools
import pytest
import numpy as np
import biotite.sequence as seq
import biotite.sequence.align as align
from .util import sequences


@pytest.mark.parametrize(
    "gap_penalty, local, band_width", itertools.product(
    [-10, (-10,-1)],
    [False, True],
    [2, 5, 20, 100]
))
def test_simple_alignment(gap_penalty, local, band_width):
    """
    Test `align_banded()` by comparing the output to `align_optimal()`.
    This test uses a pair of highly similar short sequences.
    """
    # Cyclotide C, Uniprot: P86843
    seq1 = seq.ProteinSequence("gvpcaescvwipctvtallgcsckdkvcyld")
    # Cyclotide F, Uniprot: P86846
    seq2 = seq.ProteinSequence("gipcgescvfipcissvvgcsckskvcyld")
    matrix = align.SubstitutionMatrix.std_protein_matrix()

    ref_alignments = align.align_optimal(
        seq1, seq2, matrix,
        gap_penalty=gap_penalty, local=local, terminal_penalty=False
    )
    # Remove terminal gaps in reference to obtain a true semi-global
    # alignment, as returned by align_banded()
    ref_alignments = [align.remove_terminal_gaps(al) for al in ref_alignments]
    
    test_alignments = align.align_banded(
        seq1, seq2, matrix, (-band_width, band_width),
        gap_penalty=gap_penalty, local=local
    )

    assert len(test_alignments) == len(ref_alignments)
    for alignment in test_alignments:
        assert alignment in ref_alignments


@pytest.mark.parametrize(
    "gap_penalty, local, seq_indices", itertools.product(
    [-10, (-10,-1)],
    [False, True],
    [(i,j) for i in range(10) for j in range(i+1)]
))
def test_complex_alignment(sequences, gap_penalty, local, seq_indices):
    """
    Test `align_banded()` by comparing the output to `align_optimal()`.
    This test uses a set of long sequences, which are pairwise compared.
    The band should be chosen sufficiently large so `align_banded()`
    can return the optimal alignment(s).
    """
    MAX_NUMBER = 100
    
    matrix = align.SubstitutionMatrix.std_protein_matrix()
    index1, index2 = seq_indices
    seq1 = sequences[index1]
    seq2 = sequences[index2]

    ref_alignments = align.align_optimal(
        seq1, seq2, matrix,
        gap_penalty=gap_penalty, local=local, terminal_penalty=False,
        max_number=MAX_NUMBER
    )
    # Remove terminal gaps in reference to obtain a true semi-global
    # alignment, as returned by align_banded()
    ref_alignments = [align.remove_terminal_gaps(al) for al in ref_alignments]
    
    identity = align.get_sequence_identity(ref_alignments[0])
    # Use a relatively small band width, if the sequences are similar,
    # otherwise use the entire search space
    band_width = 100 if identity > 0.5 else len(seq1) + len(seq2)
    test_alignments = align.align_banded(
        seq1, seq2, matrix, (-band_width, band_width),
        gap_penalty=gap_penalty, local=local, max_number=MAX_NUMBER
    )

    try:
        assert test_alignments[0].score == ref_alignments[0].score
        if len(ref_alignments) < MAX_NUMBER:
            # Only test if the exact same alignments were created,
            # if the number of traces was not limited by MAX_NUMBER
            assert len(test_alignments) == len(ref_alignments)
            for alignment in test_alignments:
                assert alignment in ref_alignments
    except AssertionError:
        print("First tested alignment:")
        print()
        print(test_alignments[0])
        print("\n")
        print("First reference alignment:")
        print()
        print(ref_alignments[0])
        raise


@pytest.mark.parametrize(
    "length, excerpt_length, seed", itertools.product(
    [1_000, 1_000_000],
    [50, 500],
    range(10)
))
def test_large_sequence_mapping(length, excerpt_length, seed):
    """
    Test whether an excerpt of a very large sequence is aligned to that
    sequence at the position, where the excerpt was taken from.
    """
    BAND_WIDTH = 100
    
    np.random.seed(seed)

    sequence = seq.NucleotideSequence()
    sequence.code = np.random.randint(len(sequence.alphabet), size=length)
    excerpt_pos = np.random.randint(len(sequence) - excerpt_length)
    excerpt = sequence[excerpt_pos : excerpt_pos + excerpt_length]

    diagonal = np.random.randint(
        excerpt_pos - BAND_WIDTH,
        excerpt_pos + BAND_WIDTH
    )
    band = (
        diagonal - BAND_WIDTH,
        diagonal + BAND_WIDTH
    )
    print(band)
    print(len(sequence), len(excerpt))

    matrix = align.SubstitutionMatrix.std_nucleotide_matrix()
    test_alignments = align.align_banded(
        excerpt, sequence, matrix, band=band
    )
    # The excerpt should be uniquely mappable to a single location on
    # the long sequence
    assert len(test_alignments) == 1
    test_alignment = test_alignments[0]
    test_trace = test_alignment.trace

    ref_trace = np.stack([
        np.arange(len(excerpt)),
        np.arange(excerpt_pos, len(excerpt) + excerpt_pos)
    ], axis=1)
    assert np.array_equal(test_trace, ref_trace)
    


@pytest.mark.parametrize(
    "gap_penalty, local, seed", itertools.product(
    [-10, (-10, -1)],
    [False, True],
    range(100)
))
def test_swapping(gap_penalty, local, seed):
    """
    Check if `align_banded()` returns a 'swapped' alignment, if
    the order of input sequences is swapped.
    """
    np.random.seed(seed)
    band = (
        np.random.randint(-30, -10),
        np.random.randint( 10,  30)
    )

    seq1, seq2 = _create_random_pair(seed)
    matrix = align.SubstitutionMatrix.std_protein_matrix()
    ref_alignments = align.align_banded(
        seq1, seq2, matrix, band=band, local=local, gap_penalty=gap_penalty
    )

    test_alignments = align.align_banded(
        seq2, seq1, matrix, band=band, local=local, gap_penalty=gap_penalty
    )

    if len(ref_alignments) != 1 or  len(test_alignments) != 1:
        # If multiple optimal alignments exist,
        # it is not easy to assign a swapped one to an original one
        # therefore, simply return in this case
        # the number of tested seeds should be large enough to generate
        # a reasonable number of suitable test cases
        return
    ref_alignment = ref_alignments[0]
    test_alignment = test_alignments[0]
    
    assert test_alignment.sequences[0] == ref_alignment.sequences[1]
    assert test_alignment.sequences[1] == ref_alignment.sequences[0]
    assert np.array_equal(test_alignment.trace, ref_alignment.trace[:, ::-1])



def _create_random_pair(seed, length=100, max_subsitutions=5,
                        max_insertions=5, max_deletions=5,
                        max_truncations=5):
    """
    generate a pair of protein sequences.
    Each pair contains

        1. a randomly generated sequence
        2. a sequence created by randomly introducing
           deletions/insertions/substitutions into the first sequence.
    """
    np.random.seed(seed)

    original = seq.ProteinSequence()
    original.code = np.random.randint(len(original.alphabet), size=length)

    mutant = original.copy()

    # Random Substitutions
    n_subsitutions = np.random.randint(max_subsitutions)
    subsitution_indices = np.random.choice(
        np.arange(len(mutant)), size=n_subsitutions, replace=False
    )
    subsitution_values = np.random.randint(
        len(original.alphabet), size=n_subsitutions
    )
    mutant.code[subsitution_indices] = subsitution_values

    # Random insertions
    n_insertions = np.random.randint(max_insertions)
    insertion_indices = np.random.choice(
        np.arange(len(mutant)), size=n_insertions, replace=False
    )
    insertion_values = np.random.randint(
        len(original.alphabet), size=n_insertions
    )
    mutant.code = np.insert(mutant.code, insertion_indices, insertion_values)

    # Random deletions
    n_deletions = np.random.randint(max_deletions)
    deletion_indices = np.random.choice(
        np.arange(len(mutant)), size=n_deletions, replace=False
    )
    mutant.code = np.delete(mutant.code, deletion_indices)

    # Truncate at both ends of original and mutant
    original = original[
        np.random.randint(max_truncations) :
        -(1 + np.random.randint(max_truncations))
    ]
    mutant = mutant[
        np.random.randint(max_truncations) :
        -(1 + np.random.randint(max_truncations))
    ]

    return original, mutant


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This subpackage is used for reading and writing an :class:`AtomArray` or
:class:`AtomArrayStack` using the binary MMTF format. This format
features a smaller file size and a highly increased I/O operation
performance, than the text based file formats.
"""

__name__ = "biotite.structure.io.mmtf"
__author__ = "Patrick Kunzmann"

from .file import *
from .convertfile import *
from .convertarray import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from tempfile import NamedTemporaryFile
import itertools
import glob
from os.path import join
import numpy as np
import numpy.random as random
import pytest
import biotite.structure as struc
import biotite.structure.io as strucio
import biotite.structure.io.mmtf as mmtf
from ..util import data_dir, cannot_import


def test_distance():
    coord1 = struc.coord([0,1,1])
    coord2 = struc.coord([0,2,2])
    assert struc.distance(coord1, coord2) == pytest.approx(np.sqrt(2))


def test_centroid():
    coord = struc.coord([[1,1,1],[0,-1,-1],[-1,0,0]])
    assert struc.centroid(coord).tolist() == [0,0,0]


def test_angle():
    coord1 = struc.coord([0,0,1])
    coord2 = struc.coord([0,0,0])
    coord3 = struc.coord([0,1,1])
    assert struc.angle(coord1, coord2, coord3) == pytest.approx(0.25*np.pi)


def test_dihedral():
    coord1 = struc.coord([-0.5,-1,0])
    coord2 = struc.coord([0,0,0])
    coord3 = struc.coord([1,0,0])
    coord4 = struc.coord([0,0,-1])
    assert struc.dihedral(coord1, coord2, coord3, coord4) \
           == pytest.approx(0.5*np.pi)


@pytest.mark.parametrize("multiple_chains", [False, True])
def test_dihedral_backbone_general(multiple_chains):
    stack = strucio.load_structure(join(data_dir("structure"), "1l2y.mmtf"))
    n_models = stack.stack_depth()
    n_res = stack.res_id[-1]
    if multiple_chains:
        stack2 = stack.copy()
        stack2.chain_id[:] = "B"
        stack = stack + stack2
        n_res = n_res * 2
    array = stack[0]
    # Test array
    phi, psi, omega = struc.dihedral_backbone(array)
    assert   phi.shape == (n_res,)
    assert   psi.shape == (n_res,)
    assert omega.shape == (n_res,)
    _assert_plausible_omega(omega)
    # Test stack
    phi, psi, omega = struc.dihedral_backbone(stack)
    assert   phi.shape == (n_models, n_res)
    assert   psi.shape == (n_models, n_res)
    assert omega.shape == (n_models, n_res)
    _assert_plausible_omega(omega)

def _assert_plausible_omega(omega):
    # Remove nan values
    omega = omega.flatten()
    omega = np.abs(omega[~np.isnan(omega)])
    # Omega should be always approximately pi
    assert omega.tolist() == pytest.approx([np.pi] * len(omega), rel=0.6)


@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
@pytest.mark.parametrize(
    "file_name", glob.glob(join(data_dir("structure"), "*.mmtf"))
)
def test_dihedral_backbone_result(file_name):
    import mdtraj
    
    mmtf_file = mmtf.MMTFFile.read(file_name)
    array = mmtf.get_structure(mmtf_file, model=1)
    array = array[struc.filter_amino_acids(array)]
    if array.array_length() == 0:
        # Structure contains no protein
        # -> determination of backbone angles makes no sense
        return
    
    for chain in struc.chain_iter(array):
        print("Chain: ", chain.chain_id[0])
        if len(struc.check_res_id_continuity(chain)) != 0:
            # Do not test discontinuous chains
            return
        test_phi, test_psi, test_ome = struc.dihedral_backbone(chain)

        temp = NamedTemporaryFile("w+", suffix=".pdb")
        strucio.save_structure(temp.name, chain)
        traj = mdtraj.load(temp.name)
        temp.close()
        _, ref_phi = mdtraj.compute_phi(traj)
        _, ref_psi = mdtraj.compute_psi(traj)
        _, ref_ome = mdtraj.compute_omega(traj)
        ref_phi, ref_psi, ref_ome = ref_phi[0], ref_psi[0], ref_ome[0]

        assert test_phi[1: ] == pytest.approx(ref_phi, abs=1e-5, rel=5e-3)
        assert test_psi[:-1] == pytest.approx(ref_psi, abs=1e-5, rel=5e-3)
        assert test_ome[:-1] == pytest.approx(ref_ome, abs=1e-5, rel=5e-3)



def test_index_distance_non_periodic():
    """
    Without PBC the result should be equal to the normal distance
    calculation.
    """
    array = strucio.load_structure(join(data_dir("structure"), "3o5r.mmtf"))
    ref_dist = struc.distance(
        array.coord[np.newaxis, :, :],
        array.coord[:, np.newaxis, :]
    ).flatten()
    length = array.array_length()
    dist = struc.index_distance(
        array,
        indices = np.stack([
            np.repeat(np.arange(length), length),
              np.tile(np.arange(length), length)
        ], axis=1)
    )
    assert np.allclose(dist, ref_dist)


@pytest.mark.parametrize(
    "shift", [
        np.array([10, 20, 30]),
        np.array([-8, 12, 28]),
        np.array([ 0, 99, 54])
    ]
)
def test_index_distance_periodic_orthogonal(shift):
    """
    The PBC aware computation, should give the same results,
    irrespective of which atoms are centered in the box 
    """
    array = strucio.load_structure(join(data_dir("structure"), "3o5r.mmtf"))
    # Use a box based on the boundaries of the structure
    # '+1' to add a margin
    array.box = np.diag(
        np.max(array.coord, axis=0) - np.min(array.coord, axis=0) + 1
    )

    length = array.array_length()
    dist_indices = np.stack([
        np.repeat(np.arange(length), length),
            np.tile(np.arange(length), length)
    ], axis=1)
    ref_dist = struc.index_distance(array, dist_indices, periodic=True)
    
    array.coord += shift
    array.coord = struc.move_inside_box(array.coord, array.box)
    dist = struc.index_distance(array, dist_indices, periodic=True)
    assert np.allclose(dist, ref_dist, atol=1e-5)


@pytest.mark.filterwarnings("ignore")
@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
@pytest.mark.parametrize(
    "shift, angles", itertools.product(
    [
        np.array([10, 20, 30]),
        np.array([-8, 12, 28]),
        np.array([ 0, 99, 54])
    ],
    [
        np.array([ 50,  90,  90]),
        np.array([ 90,  90, 120]),
        np.array([ 60,  60,  60])
    ]
    )
)
def test_index_distance_periodic_triclinic(shift, angles):
    """
    The PBC aware computation, should give the same results,
    irrespective of which atoms are centered in the box 
    """
    array = strucio.load_structure(join(data_dir("structure"), "3o5r.mmtf"))
    # Use a box based on the boundaries of the structure
    # '+1' to add a margin
    boundaries = np.max(array.coord, axis=0) - np.min(array.coord, axis=0) + 1
    angles = np.deg2rad(angles)
    array.box = struc.vectors_from_unitcell(
        boundaries[0], boundaries[1], boundaries[2],
        angles[0], angles[1], angles[2]
    )

    length = array.array_length()
    dist_indices = np.stack([
        np.repeat(np.arange(length), length),
          np.tile(np.arange(length), length)
    ], axis=1)
    # index_distance() creates a large ndarray
    try:
        ref_dist = struc.index_distance(array, dist_indices, periodic=True)
    except MemoryError:
        pytest.skip("Not enough memory")

    # Compare with MDTraj
    import mdtraj
    traj = mdtraj.load(join(data_dir("structure"), "3o5r.pdb"))
    # Angstrom to Nanometers
    traj.unitcell_vectors = array.box[np.newaxis, :, :] / 10
    # Nanometers to Angstrom
    mdtraj_dist = mdtraj.compute_distances(traj, dist_indices)[0] * 10
    ind = np.where(~np.isclose(ref_dist, mdtraj_dist, atol=1e-5, rtol=1e-3))[0]
    assert np.allclose(ref_dist, mdtraj_dist, atol=1e-5, rtol=1e-3)
    
    # Compare with shifted variant
    array.coord += shift
    array.coord = struc.move_inside_box(array.coord, array.box)
    # index_distance() creates a large ndarray
    try:
        test_dist = struc.index_distance(array, dist_indices, periodic=True)
    except MemoryError:
        pytest.skip("Not enough memory")
    assert np.allclose(test_dist, ref_dist, atol=1e-5)


def test_index_functions():
    """
    The `index_xxx()` functions should give the same result as the
    corresponding `xxx` functions.
    """
    stack = strucio.load_structure(join(data_dir("structure"), "1l2y.mmtf"))
    array = stack[0]
    # Test for atom array, stack and raw coordinates
    samples = (array, stack, struc.coord(array), struc.coord(stack))
    # Generate random indices
    random.seed(42)
    indices = random.randint(array.array_length(), size=(100,4), dtype=int)
    for sample in samples:
        if isinstance(sample, np.ndarray):
            atoms1 = sample[..., indices[:,0], :]
            atoms2 = sample[..., indices[:,1], :]
            atoms3 = sample[..., indices[:,2], :]
            atoms4 = sample[..., indices[:,3], :]
        else:
            atoms1 = sample[..., indices[:,0]]
            atoms2 = sample[..., indices[:,1]]
            atoms3 = sample[..., indices[:,2]]
            atoms4 = sample[..., indices[:,3]]
        assert np.allclose(
            struc.displacement(atoms1, atoms2),
            struc.index_displacement(sample, indices[:,:2]),
            atol=1e-5
        )
        assert np.allclose(
            struc.distance(atoms1, atoms2),
            struc.index_distance(sample, indices[:,:2]),
            atol=1e-5
        )
        assert np.allclose(
            struc.angle(atoms1, atoms2, atoms3),
            struc.index_angle(sample, indices[:,:3]),
            atol=1e-5
        )
        assert np.allclose(
            struc.dihedral(atoms1, atoms2, atoms3, atoms4),
            struc.index_dihedral(sample, indices[:,:4]),
            atol=1e-5
        )






# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.info"
__author__ = "Patrick Kunzmann"
__all__ = ["residue"]

from os.path import join, dirname, realpath
import msgpack
import numpy as np
from ..atoms import AtomArray
from ..bonds import BondList


_residues = None


def _init_dataset():
    """
    Load the residue dataset from MessagePack file.

    Since loading the database is computationally expensive,
    this is only done, when the residue database is actually required.
    """
    global _residues
    if _residues is not None:
        # Database is already initialized
        return

    # Residue data is taken from
    # ftp://ftp.wwpdb.org/pub/pdb/data/monomers/components.cif
    # (2019/01/27)
    _info_dir = dirname(realpath(__file__))
    with open(join(_info_dir, "residues.msgpack"), "rb") as file:
        _residues = msgpack.unpack(
            file, use_list=False, raw=False
        )


def residue(res_name):
    """
    Get an atom array, representing the residue with the given name.

    This atom array includes proper values for the ``residue_name``,
    ``hetero``, ``atom_name``, ``element`` and ``charge`` annotation
    arrays and bonds and coordinates.

    Parameters
    ----------
    res_name : str
        The up to 3-letter name of the residue.

    Returns
    -------
    atom_array : AtomArray
        The atom array described by `res_name`.

    Examples
    --------

    >>> alanine = residue("ALA")
    >>> # Atoms and geometry
    >>> print(alanine)
                0  ALA N      N        -0.966    0.493    1.500
                0  ALA CA     C         0.257    0.418    0.692
                0  ALA C      C        -0.094    0.017   -0.716
                0  ALA O      O        -1.056   -0.682   -0.923
                0  ALA CB     C         1.204   -0.620    1.296
                0  ALA OXT    O         0.661    0.439   -1.742
                0  ALA H      H        -1.383   -0.425    1.482
                0  ALA H2     H        -0.676    0.661    2.452
                0  ALA HA     H         0.746    1.392    0.682
                0  ALA HB1    H         1.459   -0.330    2.316
                0  ALA HB2    H         0.715   -1.594    1.307
                0  ALA HB3    H         2.113   -0.676    0.697
                0  ALA HXT    H         0.435    0.182   -2.647
    >>> # Bonds
    >>> print(alanine.atom_name[alanine.bonds.as_array()[:,:2]])
    [['N' 'CA']
     ['N' 'H']
     ['N' 'H2']
     ['CA' 'C']
     ['CA' 'CB']
     ['CA' 'HA']
     ['C' 'O']
     ['C' 'OXT']
     ['CB' 'HB1']
     ['CB' 'HB2']
     ['CB' 'HB3']
     ['OXT' 'HXT']]
    """
    _init_dataset()
    array_dict = _residues[res_name]

    array = AtomArray(len(array_dict["res_name"]))

    array.add_annotation("charge", int)

    array.res_name = array_dict["res_name"]
    array.atom_name = array_dict["atom_name"]
    array.element = array_dict["element"]
    array.charge = array_dict["charge"]
    array.hetero = array_dict["hetero"]

    array.coord[:,0] = array_dict["coord_x"]
    array.coord[:,1] = array_dict["coord_y"]
    array.coord[:,2] = array_dict["coord_z"]

    array.bonds = BondList(
        array.array_length(),
        bonds = np.stack([
            array_dict["bond_i"],
            array_dict["bond_j"],
            array_dict["bond_type"]
        ]).T
    )

    return array


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
A subpackage for protein secondary structure annotation using DSSP.
"""

__name__ = "biotite.application.dssp"
__author__ = "Patrick Kunzmann"

from .app import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import warnings
import numpy as np
from .seqtypes import NucleotideSequence, ProteinSequence, GeneralSequence
from .alphabet import LetterAlphabet
from .align.alignment import get_codes

__name__ = "biotite.sequence"
__author__ = "Maximilian Greil"
__all__ = ["SequenceProfile"]

# Abbreviations
_NUC_DNA_ALPH = NucleotideSequence.alphabet_unamb
_NUC_RNA_ALPH = LetterAlphabet(["A", "C", "G", "U"])
_PROT_ALPH = ProteinSequence.alphabet


def _determine_common_alphabet(alphabets):
    """
    Determine the common alphabet from a list of alphabets, that
    extends all alphabets.
    """
    common_alphabet = alphabets[0]
    for alphabet in alphabets[1:]:
        if not common_alphabet.extends(alphabet):
            if alphabet.extends(common_alphabet):
                common_alphabet = alphabet
            else:
                raise ValueError(
                    "There is no common alphabet that extends all alphabets"
                )
    return common_alphabet


def _codes_to_iupac(frequency, codes, maxes, row):
    """
    Returns IUPAC code for a row of 'symbols' with none, one or
    multiple maximum positions.
    """
    if np.sum(frequency) == 0:
        raise ValueError(
            f"There is an empty column in the 'symbols' frequency table. "
            f"This doesn't make sense in context of an alignment. "
            f"Please check the 'symbols' frequency table in row {row}."
        )
    key = tuple(np.where(frequency == maxes)[0])
    return codes[key]


class SequenceProfile(object):
    """
    A :class:`SequenceProfile` object stores information about a
    sequence profile of aligned sequences.
    It is possible to calculate and return its consensus sequence.

    This class saves the position frequency matrix
    (position count matrix) 'symbols' of the occurrences of each
    alphabet symbol at each position.
    It also saves the number of gaps at each position in the array
    'gaps'.

    With :meth:`probability_matrix()` the position probability matrix
    can be created based on 'symbols' and a pseudocount.

    With :meth:`log_odds_matrix()` the position weight matrix can
    be created based on the before calculated position probability
    matrix and the background frequencies.

    With :meth:`from_alignment()` a :class:`SequenceProfile` object can
    be created from an indefinite number of aligned sequences.

    With :meth:`sequence_probability_from_matrix()` the probability of a
    sequence can be calculated based on the before calculated position 
    probability matrix of this instance of object SequenceProfile.

    With :meth:`sequence_score_from_matrix()` the score of a sequence
    can be calculated based on the before calculated position weight
    matrix of this instance of object SequenceProfile.

    All attributes of this class are publicly accessible.

    Parameters
    ----------
    symbols : ndarray, dtype=int, shape=(n,k)
        This matrix simply saves for each position how often absolutely
        each symbol is present.
    gaps : ndarray, dtype=int, shape=n
        Array which indicates the number of gaps at each position.
    alphabet : Alphabet, length=k
        Alphabet of sequences of sequence profile

    Attributes
    ----------
    symbols : ndarray, dtype=int, shape=(n,k)
        This matrix simply saves for each position how often absolutely
        each symbol is present.
    gaps : ndarray, dtype=int, shape=n
        Array which indicates the number of gaps at each position.
    alphabet : Alphabet, length=k
        Alphabet of sequences of sequence profile
    """

    def __init__(self, symbols, gaps, alphabet):
        self._symbols = symbols
        self._gaps = gaps
        self._alphabet = alphabet

        if len(alphabet) != symbols.shape[1]:
            raise ValueError(
                f"The given alphabet doesn't have the same length "
                f"({len(alphabet)}) as the number of columns "
                f"({symbols.shape[1]}) in the 'symbols' frequency table."
            )

        if gaps.shape[0] != symbols.shape[0]:
            raise ValueError(
                f"The given 'gaps' position matrix doesn't have the same "
                f"length ({gaps.shape[0]}) as the 'symbols' "
                f"frequency table ({symbols.shape[0]})"
            )

    @property
    def symbols(self):
        return self._symbols

    @property
    def gaps(self):
        return self._gaps

    @property
    def alphabet(self):
        return self._alphabet

    @symbols.setter
    def symbols(self, new_symbols):
        if not new_symbols.shape == self.symbols.shape:
            raise ValueError(
                f"New ndarray 'symbols' must be of same shape "
                f"{self.symbols.shape} as the old one"
            )
        self._symbols = new_symbols

    @gaps.setter
    def gaps(self, new_gaps):
        if not new_gaps.shape == self.gaps.shape:
            raise ValueError(
                f"New ndarray 'gaps' must be of same shape "
                f"{self.gaps.shape} as the old one"
            )
        self._gaps = new_gaps

    def __repr__(self):
        """Represent SequenceProfile as a string for debugging."""
        return f"SequenceProfile(np.{np.array_repr(self.symbols)}, " \
               f"np.{np.array_repr(self.gaps)}, Alphabet({self.alphabet}))"

    def __eq__(self, item):
        if not isinstance(item, SequenceProfile):
            return False
        if not np.array_equal(self.symbols, item.symbols):
            return False
        if not np.array_equal(self.gaps, item.gaps):
            return False
        if not self.alphabet == item.alphabet:
            return False
        return True

    @staticmethod
    def from_alignment(alignment, alphabet=None):
        """
        Get an object of :class:`SequenceProfile` from an object of
        :class:`Alignment`.

        Based on the sequences of the alignment, the SequenceProfile
        parameters symbols and gaps are calculated.

        Parameters
        ----------
        alignment : Alignment
            An Alignment object to create the SequenceProfile object
            from.
        alphabet : bool
            This alphabet will be used when creating the SequenceProfile
            object. If no alphabet is selected, the alphabet for this
            SequenceProfile
            object will be calculated from the sequences of object
            Alignment.
            (Default: None).

        Returns
        -------
        profile: SequenceProfile
            The created SequenceProfile object
        """
        sequences = get_codes(alignment)
        if alphabet is None:
            alphabet = _determine_common_alphabet(
                [seq.alphabet for seq in alignment.sequences]
            )
        else:
            for alph in (seq.alphabet for seq in alignment.sequences):
                if not alphabet.extends(alph):
                    raise ValueError(
                        f"The given alphabet is incompatible with a least one "
                        "alphabet of the given sequences"
                    )
        symbols = np.zeros((len(sequences[0]), len(alphabet)), dtype=int)
        gaps = np.zeros(len(sequences[0]), dtype=int)
        sequences = np.transpose(sequences)
        for i in range(len(sequences)):
            row = np.where(sequences[i, ] == -1, len(alphabet), sequences[i, ])
            count = np.bincount(row, minlength=len(alphabet) + 1)
            symbols[i, ] = count[0:len(alphabet)]
            gaps[i] = count[-1]
        return SequenceProfile(symbols, gaps, alphabet)

    def to_consensus(self, as_general=False):
        """
        Get the consensus sequence for this SequenceProfile object.

        Parameters
        ----------
        as_general : bool
            If true, returns consensus sequence as GeneralSequence
            object.
            Otherwise, the consensus sequence object type is chosen
            based on the alphabet of this SequenceProfile object
            (Default: False).

        Returns
        -------
        consensus: Sequence
            The calculated consensus sequence
        """
        # https://en.wikipedia.org/wiki/International_Union_of_Pure_and_Applied_Chemistry#Amino_acid_and_nucleotide_base_codes
        if as_general:
            return self._general_to_consensus()
        elif self.alphabet == _NUC_DNA_ALPH:
            return NucleotideSequence(self._dna_to_consensus())
        elif self.alphabet == _NUC_RNA_ALPH:
            return NucleotideSequence(self._rna_to_consensus())
        elif self.alphabet == _PROT_ALPH:
            return self._prot_to_consensus()
        return self._general_to_consensus()

    def _dna_to_consensus(self):
        codes = {
            (0,): 'A', (1,): 'C', (2,): 'G', (3,): 'T',
            (0, 2): 'R', (1, 3): 'Y', (1, 2): 'S', (0, 3): 'W', (2, 3): 'K', (0, 1): 'M',
            (1, 2, 3): 'B', (0, 2, 3): 'D', (0, 1, 3): 'H', (0, 1, 2): 'V',
            (0, 1, 2, 3): 'N'
        }
        consensus = ""
        maxes = np.max(self.symbols, axis=1)
        for i in range(len(self.symbols)):
            consensus += _codes_to_iupac(self.symbols[i, :], codes, maxes[i], i)
        return consensus

    def _rna_to_consensus(self):
        codes = {
            (0,): 'A', (1,): 'C', (2,): 'G', (3,): 'U',
            (0, 2): 'R', (1, 3): 'Y', (1, 2): 'S', (0, 3): 'W', (2, 3): 'K', (0, 1): 'M',
            (1, 2, 3): 'B', (0, 2, 3): 'D', (0, 1, 3): 'H', (0, 1, 2): 'V',
            (0, 1, 2, 3): 'N'
        }
        consensus = ""
        maxes = np.max(self.symbols, axis=1)
        for i in range(len(self.symbols)):
            consensus += _codes_to_iupac(self.symbols[i, :], codes, maxes[i], i)
        return consensus

    def _prot_to_consensus(self):
        """
        In case there is more than one symbol with the same maximal
        occurrences, the alphabetically sorted first symbol will be
        taken for the consensus sequence.
        """
        consensus = ProteinSequence()
        consensus.code = np.argmax(self.symbols, axis=1)
        consensus.code = np.where(
            np.sum(self.symbols, axis=1) == 0, 23, consensus.code
        )  # _PROT_ALPH[23] = 'X'
        return consensus

    def _general_to_consensus(self):
        """
        In case there is more than one symbol with the same maximal
        occurrences, the alphabetically sorted first symbol will be
        taken for the consensus sequence.
        In case the sum of occurrences of all symbols at a position is
        zero, the alphabetically sorted first symbol will be taken for
        the consensus sequence.
        """
        consensus = GeneralSequence(self.alphabet)
        consensus.code = np.argmax(self.symbols, axis=1)
        return consensus

    def probability_matrix(self, pseudocount=0):
        r"""
        Calculate the position probability matrix (PPM) based on
        'symbols' and the given pseudocount.
        This new matrix has the same shape as 'symbols'.

        .. math::

            P(S) = \frac {C_S + \frac{c_p}{k}} {\sum_{i} C_i + c_p}
        
        :math:`S`: The symbol.

        :math:`C_S`: The count of symbol :math:`S` at the sequence
        position.

        :math:`c_p`: The pseudocount.

        :math:`k`: Length of the alphabet.

        Parameters
        ----------
        pseudocount: int, optional
            Amount added to the number of observed cases in order to
            change the expected probability of the PPM.
            (Default: 0)

        Returns
        -------
        ppm: ndarray, dtype=float, shape=(n,k)
            The calculated the position probability matrix.
        """
        if pseudocount < 0:
            raise ValueError(
                f"Pseudocount can not be smaller than zero."
            )
        return (self.symbols + pseudocount / self.symbols.shape[1]) / \
               (np.sum(self.symbols, axis=1)[:, np.newaxis] + pseudocount)

    def log_odds_matrix(self, background_frequencies=None, pseudocount=0):
        r"""
        Calculate the position weight matrix (PWM) based on the
        position probability matrix (PPM) (with given pseudocount) and
        background_frequencies.
        This new matrix has the same shape as 'symbols'.

        .. math::

            W(S) = \log_2 \left( \frac{P(S)}{B_S} \right)
        
        :math:`S`: The symbol.

        :math:`P(S)`: The probability of symbol :math:`S` at the
        sequence position.

        :math:`c_p`: The background frequency of symbol :math:`S`.

        Parameters
        ----------
        pseudocount: int, optional
            Amount added to the number of observed cases in order to change
            the expected probability of the PPM.
            (Default: 0)
        background_frequencies: ndarray, shape=(k,), dtype=float, optional
            The background frequencies for each symbol in the alphabet.
            By default, a uniform distribution is assumed.
            
        Returns
        -------
        pwm: ndarray, dtype=float, shape=(n,k)
            The calculated the position weight matrix.
        """
        if background_frequencies is None:
            background_frequencies = 1 / len(self.alphabet)
        ppm = self.probability_matrix(pseudocount=pseudocount)
        # Catch warning that appears, if a symbol is missing at any
        # position in the profile
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            return np.log2(ppm / background_frequencies)

    def sequence_probability(self, sequence, pseudocount=0):
        r"""
        Calculate probability of a sequence based on the
        position probability matrix (PPM).

        The sequence probability is the product of the probability of 
        the respective symbol over all sequence positions.

        Parameters
        ----------
        sequence : Sequence
           The input sequence.
        pseudocount: int, optional
            Amount added to the number of observed cases in order to change
            the expected probability of the PPM.
            (Default: 0)

        Returns
        -------
        probability: float
           The calculated probability for the input sequence based on
           the PPM.
        """
        ppm = self.probability_matrix(pseudocount=pseudocount)
        if len(sequence) != len(ppm):
            raise ValueError(
                f"The given sequence has a different length ({len(sequence)}) than "
                f"the position probability matrix ({len(ppm)})."
            )
        if not ppm.shape == self.symbols.shape:
            raise ValueError(
                f"Position probability matrix {ppm.shape} must be of same shape "
                f"as 'symbols' {self.symbols.shape}"
            )
        return np.prod(ppm[np.arange(len(sequence)), sequence.code])

    def sequence_score(self, sequence, background_frequencies=None, pseudocount=0):
        """
        Calculate score of a sequence based on the
        position weight matrix (PWM).

        The score is the sum of weights (log-odds scores) of 
        the respective symbol over all sequence positions.

        Parameters
        ----------
        sequence : Sequence
           The input sequence.
        pseudocount: int, optional
            Amount added to the number of observed cases in order to change
            the expected probability of the PPM.
            (Default: 0)
        background_frequencies: ndarray, shape=(k,), dtype=float, optional
            The background frequencies for each symbol in the alphabet.
            By default a uniform distribution is assumed.

        Returns
        -------
        score: float
           The calculated score for the input sequence based on
           the PWM.
        """
        if background_frequencies is None:
            background_frequencies = 1 / len(self.alphabet)
        pwm = self.log_odds_matrix(background_frequencies=background_frequencies, pseudocount=pseudocount)
        if len(sequence) != len(pwm):
            raise ValueError(
                f"The given sequence has a different length ({len(sequence)}) than "
                f"the position weight matrix ({len(pwm)})."
            )
        if not pwm.shape == self.symbols.shape:
            raise ValueError(
                f"Position weight matrix {pwm.shape} must be of same shape "
                f"as 'symbols' {self.symbols.shape}"
            )
        return np.sum(pwm[np.arange(len(sequence)), sequence.code])



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite"
__author__ = "Patrick Kunzmann"
__all__ = ["File", "TextFile", "InvalidFileError"]

import abc
import io
import warnings
from .copyable import Copyable
import copy


class File(Copyable, metaclass=abc.ABCMeta):
    """
    Base class for all file classes.
    The constructor creates an empty file, that can be filled with data
    using the class specific setter methods.
    Conversely, the class method :func:`read()` reads a file from disk
    (or a file-like object from other sources).
    In order to write the instance content into a file the
    :func:`write()` method is used.
    """
    
    def __init__(self):
        # Support for deprecated instance method 'read()':
        # When creating an instance, the 'read()' class method is
        # replaced by the instance method, so that subsequent
        # 'read()' calls are delegated to the instance method
        self.read = self._deprecated_read
    
    @classmethod
    @abc.abstractmethod
    def read(cls, file):
        """
        Parse a file (or file-like object).
        
        Parameters
        ----------
        file : file-like object or str
            The file to be read.
            Alternatively a file path can be supplied.
        
        Returns
        -------
        file_object : File
            An instance from the respective :class:`File` subclass
            representing the parsed file.
        """
        pass
        
            
    def _deprecated_read(self, file, *args, **kwargs):
        """
        Support for deprecated instance method :func:`read()`.

        Internally this calls the :func:`read()` class method and
        replaces the data in `self` with the data from the newly created
        :class:`File` object
        """
        warnings.warn(
            "Instance method 'read()' is deprecated, "
            "use class method instead",
            DeprecationWarning
        )
        cls = type(self)
        new_file = cls.read(file, *args, **kwargs)
        self.__dict__.update(new_file.__dict__)
    
    @abc.abstractmethod
    def write(self, file):
        """
        Write the contents of this :class:`File` object into a file.
        
        Parameters
        ----------
        file_name : file-like object or str
            The file to be written to.
            Alternatively a file path can be supplied.
        """
        pass
        

class TextFile(File, metaclass=abc.ABCMeta):
    """
    Base class for all line based text files.
    When reading a file, the text content is saved as list of strings,
    one for each line.
    When writing a file, this list is written into the file.
    
    Attributes
    ----------
    lines : list
        List of string representing the lines in the text file.
        PROTECTED: Do not modify from outside.
    """
    
    def __init__(self):
        super().__init__()
        self.lines = []

    @classmethod
    def read(cls, file, *args, **kwargs):
        # File name
        if isinstance(file, str):
            with open(file, "r") as f:
                lines = f.read().splitlines()
        # File object
        else:
            if not is_text(file):
                raise TypeError("A file opened in 'text' mode is required")
            lines = file.read().splitlines()
        file_object = cls(*args, **kwargs)
        file_object.lines = lines
        return file_object
    
    @staticmethod
    def read_iter(file):
        """
        Create an iterator over each line of the given text file.
        
        Parameters
        ----------
        file : file-like object or str
            The file to be read.
            Alternatively a file path can be supplied.
        
        Yields
        ------
        line : str
            The current line in the file.
        """
        # File name
        if isinstance(file, str):
            with open(file, "r") as f:
                yield from f
        # File object
        else:
            if not is_text(file):
                raise TypeError("A file opened in 'text' mode is required")
            yield from file

    def write(self, file):
        """
        Write the contents of this object into a file
        (or file-like object).
        
        Parameters
        ----------
        file : file-like object or str
            The file to be written to.
            Alternatively a file path can be supplied.
        """
        if isinstance(file, str):
            with open(file, "w") as f:
                f.write("\n".join(self.lines) + "\n")
        else:
            if not is_text(file):
                raise TypeError("A file opened in 'text' mode is required")
            file.write("\n".join(self.lines) + "\n")
    
    @staticmethod
    def write_iter(file, lines):
        """
        Iterate over the given `lines` of text and write each line into
        the specified `file`.

        In contrast to :meth:`write()`, each line of text is not stored
        in an intermediate :class:`TextFile`, but is directly written
        to the file.
        Hence, this static method may save a large amount of memory if
        a large file should be written, especially if the `lines`
        are provided as generator.
        
        Parameters
        ----------
        file : file-like object or str
            The file to be written to.
            Alternatively a file path can be supplied.
        lines : generator or array-like of str
            The lines of text to be written.
            Must not include line break characters.
        """
        if isinstance(file, str):
            with open(file, "w") as f:
                for line in lines:
                    f.write(line + "\n")
        else:
            if not is_text(file):
                raise TypeError("A file opened in 'text' mode is required")
            for line in lines:
                file.write(line + "\n")
    
    def __copy_fill__(self, clone):
        super().__copy_fill__(clone)
        clone.lines = copy.copy(self.lines)
    
    def __str__(self):
        return("\n".join(self.lines))


class InvalidFileError(Exception):
    """
    Indicates that the file is not suitable for the requested action,
    either because the file does not contain the required data or
    because the file is malformed.
    """
    pass


def wrap_string(text, width):
    """
    A much simpler and hence much more efficient version of
    `textwrap.wrap()`.

    This function simply wraps the given `text` after `width`
    characters, ignoring sentences, whitespaces, etc.
    """
    lines = []
    for i in range(0, len(text), width):
        lines.append(text[i : i+width])
    return lines


def is_binary(file):
    if isinstance(file, io.BufferedIOBase):
        return True
    # for file wrappers, e.g. 'TemporaryFile'
    return hasattr(file, "file") and isinstance(file.file, io.BufferedIOBase)


def is_text(file):
    if isinstance(file, io.TextIOBase):
        return True
    # for file wrappers, e.g. 'TemporaryFile'
    return hasattr(file, "file") and isinstance(file.file, io.TextIOBase)



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.io.fasta"
__author__ = "Patrick Kunzmann"

import warnings
from collections import OrderedDict
from ...sequence import Sequence
from ...alphabet import AlphabetError, LetterAlphabet
from ...seqtypes import NucleotideSequence, ProteinSequence
from ...align.alignment import Alignment

__all__ = ["get_sequence", "get_sequences", "set_sequence", "set_sequences",
           "get_alignment", "set_alignment"]


def get_sequence(fasta_file, header=None):
    """
    Get a sequence from a :class:`FastaFile` instance.

    The type of sequence is guessed from the sequence string:
    First, a conversion into a :class:`NucleotideSequence` and
    second a conversion into a :class:`ProteinSequence` is tried.
    
    Parameters
    ----------
    fasta_file : FastaFile
        The :class:`FastaFile` to be accessed.
    header : str, optional
        The header to get the sequence from. By default, the first
        sequence of the file is returned.
    
    Returns
    -------
    sequence : NucleotideSequence or ProteinSequence
        The requested sequence in the `FastaFile`.
        :class:`NucleotideSequence` if the sequence string fits the
        corresponding alphabet, :class:`ProteinSequence` otherwise.
    
    Raises
    ------
    ValueError
        If the sequence data can be neither converted into a
        :class:`NucleotideSequence` nor a :class:`ProteinSequence`.
    """
    if header is not None:
        seq_str = fasta_file[header]
    else:
        # Return first (and probably only) sequence of file
        seq_str = None
        for seq_str in fasta_file.values():
            break
        if seq_str is None:
            raise ValueError("File does not contain any sequences")
    # Determine the sequence type:
    # If NucleotideSequence can be created it is a DNA sequence,
    # otherwise protein sequence
    return _convert_to_sequence(seq_str)


def get_sequences(fasta_file):
    """
    Get dictionary from a :class:`FastaFile` instance,
    where headers are keys and sequences are values.

    The type of sequence is guessed from the sequence string:
    First, a conversion into a :class:`NucleotideSequence` and
    second a conversion into a :class:`ProteinSequence` is tried.
    
    Parameters
    ----------
    fasta_file : FastaFile
        The :class:`FastaFile` to be accessed.
    
    Returns
    -------
    seq_dict : dict
        A dictionary that maps headers to
        :class:`NucleotideSequence` and/or :class:`ProteinSequence`
        instances as values.
    
    Raises
    ------
    ValueError
        If at least on of the sequence strings can be neither converted
        into a :class:`NucleotideSequence` nor a
        :class:`ProteinSequence`.
    """
    seq_dict = OrderedDict()
    for header, seq_str in fasta_file.items():
        seq_dict[header] = _convert_to_sequence(seq_str)
    return seq_dict


def set_sequence(fasta_file, sequence, header=None, as_rna=False):
    """
    Set a sequence in a :class:`FastaFile` instance.
    
    Parameters
    ----------
    fasta_file : FastaFile
        The :class:`FastaFile` to be accessed.
    sequence : Sequence
        The sequence to be set.
    header : str, optional
        The header for the sequence. Default is ``'sequence'``.
    as_rna : bool, optional
        If set to true, ``'T'`` will be replaced by ``'U'``,
        if a :class:`NucleotideSequence` was given.
    
    Raises
    ------
    ValueError
        If the sequence's alphabet uses symbols other than single
        characters.
    """
    if header is None:
        header = "sequence"
    fasta_file[header] = _convert_to_string(sequence, as_rna)


def set_sequences(fasta_file, sequence_dict, as_rna=False):
    """
    Set sequences in a :class:`FastaFile` instance from a dictionary.
    
    Parameters
    ----------
    fasta_file : FastaFile
        The :class:`FastaFile` to be accessed.
    sequence_dict : dict
        A dictionary containing the sequences to be set.
        Header are keys, :class:`Sequence` instances are values.
    as_rna : bool, optional
        If set to true, ``'T'`` will be replaced by ``'U'``,
        if a :class:`NucleotideSequence` was given.
    
    Raises
    ------
    ValueError
        If the sequences alphabets uses symbols other than single
        characters.
    """
    for header, sequence in sequence_dict.items():
        fasta_file[header] = _convert_to_string(sequence, as_rna)


def get_alignment(fasta_file, additional_gap_chars=("_",)):
    """
    Get an alignment from a :class:`FastaFile` instance.
    
    Parameters
    ----------
    fasta_file : FastaFile
        The :class:`FastaFile` to be accessed.
    additional_gap_chars : str, optional
        The characters to be treated as gaps.
    
    Returns
    -------
    alignment : Alignment
        The alignment from the :class:`FastaFile`.
    """
    seq_strings = list(fasta_file.values())
    # Replace additional gap symbols with default gap symbol ('-')
    for char in additional_gap_chars:
        for i, seq_str in enumerate(seq_strings):
            seq_strings[i] = seq_str.replace(char, "-")
    # Remove gaps for creation of sequences
    sequences = [_convert_to_sequence(seq_str.replace("-",""))
                 for seq_str in seq_strings]
    trace = Alignment.trace_from_strings(seq_strings)
    return Alignment(sequences, trace, score=None)


def set_alignment(fasta_file, alignment, seq_names):
    """
    Fill a :class:`FastaFile` with gapped sequence strings from an
    alignment.
    
    Parameters
    ----------
    fasta_file : FastaFile
        The :class:`FastaFile` to be accessed.
    alignment : Alignment
        The alignment to be set. 
    seq_names : iterable object of str
        The names for the sequences in the alignment.
        Must have the same length as the sequence count in `alignment`.
    """
    gapped_seq_strings = alignment.get_gapped_sequences()
    if len(gapped_seq_strings) != len(seq_names):
        raise ValueError(
            f"Alignment has {len(gapped_seq_strings)} sequences, "
            f"but {len(seq_names)} names were given"
        )
    for i in range(len(gapped_seq_strings)):
        fasta_file[seq_names[i]] = gapped_seq_strings[i]


def _convert_to_sequence(seq_str):
    # Biotite alphabets for nucleotide and proteins
    # do not accept lower case letters
    seq_str = seq_str.upper()
    try:
        # For nucleotides uracil is represented by thymine and
        # there is is only one letter for completely unknown nucleotides
        return NucleotideSequence(seq_str.replace("U","T").replace("X","N"))
    except AlphabetError:
        pass
    try:
        if "U" in seq_str:
            warn = True
            seq_str = seq_str.replace("U", "C")
        else:
            warn = False
        prot_seq = ProteinSequence(seq_str)
        # Raise Warning after conversion into 'ProteinSequence'
        # to wait for potential 'AlphabetError'
        if warn:
            warnings.warn(
                "ProteinSequence objects do not support selenocysteine (U), "
                "occurrences were substituted by cysteine (C)"
            )
        return prot_seq
    except AlphabetError:
        raise ValueError("FASTA data cannot be converted either to "
                         "'NucleotideSequence' nor to 'ProteinSequence'")


def _convert_to_string(sequence, as_rna):
    if not isinstance(sequence.get_alphabet(), LetterAlphabet):
        raise ValueError("Only sequences using single letter alphabets "
                         "can be stored in a FASTA file")
    if isinstance(sequence, NucleotideSequence) and as_rna:
        return(str(sequence).replace("T", "U"))
    else:
        return(str(sequence))
